import argparse
import io
import os.path
import sys

import pytest

from ..patterns import PathFullPattern, PathPrefixPattern, FnmatchPattern, ShellPattern, RegexPattern
from ..patterns import load_exclude_file, load_pattern_file
from ..patterns import parse_pattern, PatternMatcher
from ..patterns import get_regex_from_pattern


def check_patterns(files, pattern, expected):
    """Utility for testing patterns."""
    assert all([f == os.path.normpath(f) for f in files]), "Pattern matchers expect normalized input paths"

    matched = [f for f in files if pattern.match(f)]

    assert matched == (files if expected is None else expected)


@pytest.mark.parametrize(
    "pattern, expected",
    [
        # "None" means all files, i.e. all match the given pattern
        ("/", []),
        ("/home", ["home"]),
        ("/home///", ["home"]),
        ("/./home", ["home"]),
        ("/home/user", ["home/user"]),
        ("/home/user2", ["home/user2"]),
        ("/home/user/.bashrc", ["home/user/.bashrc"]),
    ],
)
def test_patterns_full(pattern, expected):
    files = ["home", "home/user", "home/user2", "home/user/.bashrc"]

    check_patterns(files, PathFullPattern(pattern), expected)


@pytest.mark.parametrize(
    "pattern, expected",
    [
        # "None" means all files, i.e. all match the given pattern
        ("", []),
        ("relative", []),
        ("relative/path/", ["relative/path"]),
        ("relative/path", ["relative/path"]),
    ],
)
def test_patterns_full_relative(pattern, expected):
    files = ["relative/path", "relative/path2"]

    check_patterns(files, PathFullPattern(pattern), expected)


@pytest.mark.parametrize(
    "pattern, expected",
    [
        # "None" means all files, i.e. all match the given pattern
        ("/", None),
        ("/./", None),
        ("", []),
        ("/home/u", []),
        ("/home/user", ["home/user/.profile", "home/user/.bashrc"]),
        ("/etc", ["etc/server/config", "etc/server/hosts"]),
        ("///etc//////", ["etc/server/config", "etc/server/hosts"]),
        ("/./home//..//home/user2", ["home/user2/.profile", "home/user2/public_html/index.html"]),
        ("/srv", ["srv/messages", "srv/dmesg"]),
    ],
)
def test_patterns_prefix(pattern, expected):
    files = [
        "etc/server/config",
        "etc/server/hosts",
        "home",
        "home/user/.profile",
        "home/user/.bashrc",
        "home/user2/.profile",
        "home/user2/public_html/index.html",
        "srv/messages",
        "srv/dmesg",
    ]

    check_patterns(files, PathPrefixPattern(pattern), expected)


@pytest.mark.parametrize(
    "pattern, expected",
    [
        # "None" means all files, i.e. all match the given pattern
        ("", []),
        ("foo", []),
        ("relative", ["relative/path1", "relative/two"]),
        ("more", ["more/relative"]),
    ],
)
def test_patterns_prefix_relative(pattern, expected):
    files = ["relative/path1", "relative/two", "more/relative"]

    check_patterns(files, PathPrefixPattern(pattern), expected)


@pytest.mark.parametrize(
    "pattern, expected",
    [
        # "None" means all files, i.e. all match the given pattern
        ("/*", None),
        ("/./*", None),
        ("*", None),
        (
            "*/*",
            [
                "etc/server/config",
                "etc/server/hosts",
                "home/user/.profile",
                "home/user/.bashrc",
                "home/user2/.profile",
                "home/user2/public_html/index.html",
                "srv/messages",
                "srv/dmesg",
                "home/foo/.thumbnails",
                "home/foo/bar/.thumbnails",
            ],
        ),
        (
            "*///*",
            [
                "etc/server/config",
                "etc/server/hosts",
                "home/user/.profile",
                "home/user/.bashrc",
                "home/user2/.profile",
                "home/user2/public_html/index.html",
                "srv/messages",
                "srv/dmesg",
                "home/foo/.thumbnails",
                "home/foo/bar/.thumbnails",
            ],
        ),
        ("/home/u", []),
        (
            "/home/*",
            [
                "home/user/.profile",
                "home/user/.bashrc",
                "home/user2/.profile",
                "home/user2/public_html/index.html",
                "home/foo/.thumbnails",
                "home/foo/bar/.thumbnails",
            ],
        ),
        ("/home/user/*", ["home/user/.profile", "home/user/.bashrc"]),
        ("/etc/*", ["etc/server/config", "etc/server/hosts"]),
        ("*/.pr????e", ["home/user/.profile", "home/user2/.profile"]),
        ("///etc//////*", ["etc/server/config", "etc/server/hosts"]),
        ("/./home//..//home/user2/*", ["home/user2/.profile", "home/user2/public_html/index.html"]),
        ("/srv*", ["srv/messages", "srv/dmesg"]),
        ("/home/*/.thumbnails", ["home/foo/.thumbnails", "home/foo/bar/.thumbnails"]),
    ],
)
def test_patterns_fnmatch(pattern, expected):
    files = [
        "etc/server/config",
        "etc/server/hosts",
        "home",
        "home/user/.profile",
        "home/user/.bashrc",
        "home/user2/.profile",
        "home/user2/public_html/index.html",
        "srv/messages",
        "srv/dmesg",
        "home/foo/.thumbnails",
        "home/foo/bar/.thumbnails",
    ]

    check_patterns(files, FnmatchPattern(pattern), expected)


@pytest.mark.parametrize(
    "pattern, expected",
    [
        # "None" means all files, i.e. all match the given pattern
        ("*", None),
        ("**/*", None),
        ("/**/*", None),
        ("/./*", None),
        (
            "*/*",
            [
                "etc/server/config",
                "etc/server/hosts",
                "home/user/.profile",
                "home/user/.bashrc",
                "home/user2/.profile",
                "home/user2/public_html/index.html",
                "srv/messages",
                "srv/dmesg",
                "srv2/blafasel",
                "home/foo/.thumbnails",
                "home/foo/bar/.thumbnails",
            ],
        ),
        (
            "*///*",
            [
                "etc/server/config",
                "etc/server/hosts",
                "home/user/.profile",
                "home/user/.bashrc",
                "home/user2/.profile",
                "home/user2/public_html/index.html",
                "srv/messages",
                "srv/dmesg",
                "srv2/blafasel",
                "home/foo/.thumbnails",
                "home/foo/bar/.thumbnails",
            ],
        ),
        ("/home/u", []),
        (
            "/home/*",
            [
                "home/user/.profile",
                "home/user/.bashrc",
                "home/user2/.profile",
                "home/user2/public_html/index.html",
                "home/foo/.thumbnails",
                "home/foo/bar/.thumbnails",
            ],
        ),
        ("/home/user/*", ["home/user/.profile", "home/user/.bashrc"]),
        ("/etc/*/*", ["etc/server/config", "etc/server/hosts"]),
        ("/etc/**/*", ["etc/server/config", "etc/server/hosts"]),
        ("/etc/**/*/*", ["etc/server/config", "etc/server/hosts"]),
        ("*/.pr????e", []),
        ("**/.pr????e", ["home/user/.profile", "home/user2/.profile"]),
        ("///etc//////*", ["etc/server/config", "etc/server/hosts"]),
        ("/./home//..//home/user2/", ["home/user2/.profile", "home/user2/public_html/index.html"]),
        ("/./home//..//home/user2/**/*", ["home/user2/.profile", "home/user2/public_html/index.html"]),
        ("/srv*/", ["srv/messages", "srv/dmesg", "srv2/blafasel"]),
        ("/srv*", ["srv", "srv/messages", "srv/dmesg", "srv2", "srv2/blafasel"]),
        ("/srv/*", ["srv/messages", "srv/dmesg"]),
        ("/srv2/**", ["srv2", "srv2/blafasel"]),
        ("/srv2/**/", ["srv2/blafasel"]),
        ("/home/*/.thumbnails", ["home/foo/.thumbnails"]),
        ("/home/*/*/.thumbnails", ["home/foo/bar/.thumbnails"]),
    ],
)
def test_patterns_shell(pattern, expected):
    files = [
        "etc/server/config",
        "etc/server/hosts",
        "home",
        "home/user/.profile",
        "home/user/.bashrc",
        "home/user2/.profile",
        "home/user2/public_html/index.html",
        "srv",
        "srv/messages",
        "srv/dmesg",
        "srv2",
        "srv2/blafasel",
        "home/foo/.thumbnails",
        "home/foo/bar/.thumbnails",
    ]

    check_patterns(files, ShellPattern(pattern), expected)


@pytest.mark.parametrize(
    "pattern, expected",
    [
        # "None" means all files, i.e. all match the given pattern
        ("", None),
        (".*", None),
        ("^/", None),
        ("^abc$", []),
        ("^[^/]", []),
        (
            "^(?!/srv|/foo|/opt)",
            [
                "/home",
                "/home/user/.profile",
                "/home/user/.bashrc",
                "/home/user2/.profile",
                "/home/user2/public_html/index.html",
                "/home/foo/.thumbnails",
                "/home/foo/bar/.thumbnails",
            ],
        ),
    ],
)
def test_patterns_regex(pattern, expected):
    files = [
        "/srv/data",
        "/foo/bar",
        "/home",
        "/home/user/.profile",
        "/home/user/.bashrc",
        "/home/user2/.profile",
        "/home/user2/public_html/index.html",
        "/opt/log/messages.txt",
        "/opt/log/dmesg.txt",
        "/home/foo/.thumbnails",
        "/home/foo/bar/.thumbnails",
    ]

    obj = RegexPattern(pattern)
    assert str(obj) == pattern
    assert obj.pattern == pattern

    check_patterns(files, obj, expected)


def test_regex_pattern():
    # The forward slash must match the platform-specific path separator
    assert RegexPattern("^/$").match("/")
    assert RegexPattern("^/$").match(os.path.sep)
    assert not RegexPattern(r"^\\$").match("/")


def use_normalized_unicode():
    return sys.platform in ("darwin",)


def _make_test_patterns(pattern):
    return [
        PathPrefixPattern(pattern),
        FnmatchPattern(pattern),
        RegexPattern(f"^{pattern}/foo$"),
        ShellPattern(pattern),
    ]


@pytest.mark.parametrize("pattern", _make_test_patterns("b\N{LATIN SMALL LETTER A WITH ACUTE}"))
def test_composed_unicode_pattern(pattern):
    assert pattern.match("b\N{LATIN SMALL LETTER A WITH ACUTE}/foo")
    assert pattern.match("ba\N{COMBINING ACUTE ACCENT}/foo") == use_normalized_unicode()


@pytest.mark.parametrize("pattern", _make_test_patterns("ba\N{COMBINING ACUTE ACCENT}"))
def test_decomposed_unicode_pattern(pattern):
    assert pattern.match("b\N{LATIN SMALL LETTER A WITH ACUTE}/foo") == use_normalized_unicode()
    assert pattern.match("ba\N{COMBINING ACUTE ACCENT}/foo")


@pytest.mark.parametrize("pattern", _make_test_patterns(str(b"ba\x80", "latin1")))
def test_invalid_unicode_pattern(pattern):
    assert not pattern.match("ba/foo")
    assert pattern.match(str(b"ba\x80/foo", "latin1"))


@pytest.mark.parametrize(
    "lines, expected",
    [
        # "None" means all files, i.e. none excluded
        ([], None),
        (["# Comment only"], None),
        (["*"], []),
        (
            [
                "# Comment",
                "*/something00.txt",
                "  *whitespace*  ",
                # Whitespace before comment
                " #/ws*",
                # Empty line
                "",
                "# EOF",
            ],
            ["more/data", "home", " #/wsfoobar"],
        ),
        ([r"re:.*"], []),
        ([r"re:\s"], ["data/something00.txt", "more/data", "home"]),
        ([r"re:(.)(\1)"], ["more/data", "home", "\tstart/whitespace", "whitespace/end\t"]),
        (
            [
                "",
                "",
                "",
                "# This is a test with mixed pattern styles",
                # Case-insensitive pattern
                r"re:(?i)BAR|ME$",
                "",
                "*whitespace*",
                "fm:*/something00*",
            ],
            ["more/data"],
        ),
        ([r"  re:^\s  "], ["data/something00.txt", "more/data", "home", "whitespace/end\t"]),
        ([r"  re:\s$  "], ["data/something00.txt", "more/data", "home", " #/wsfoobar", "\tstart/whitespace"]),
        (["pp:./"], None),
        # leading slash is removed
        (["pp:/"], []),
        (["pp:aaabbb"], None),
        (["pp:/data", "pp: #/", "pp:\tstart", "pp:/whitespace"], ["more/data", "home"]),
        (
            ["/nomatch", "/more/*"],
            ["data/something00.txt", "home", " #/wsfoobar", "\tstart/whitespace", "whitespace/end\t"],
        ),
        # the order of exclude patterns shouldn't matter
        (
            ["/more/*", "/nomatch"],
            ["data/something00.txt", "home", " #/wsfoobar", "\tstart/whitespace", "whitespace/end\t"],
        ),
    ],
)
def test_exclude_patterns_from_file(tmpdir, lines, expected):
    files = ["data/something00.txt", "more/data", "home", " #/wsfoobar", "\tstart/whitespace", "whitespace/end\t"]

    def evaluate(filename):
        patterns = []
        load_exclude_file(open(filename), patterns)
        matcher = PatternMatcher(fallback=True)
        matcher.add_inclexcl(patterns)
        return [path for path in files if matcher.match(path)]

    exclfile = tmpdir.join("exclude.txt")

    with exclfile.open("wt") as fh:
        fh.write("\n".join(lines))

    assert evaluate(str(exclfile)) == (files if expected is None else expected)


@pytest.mark.parametrize(
    "lines, expected_roots, expected_numpatterns",
    [
        # "None" means all files, i.e. none excluded
        ([], [], 0),
        (["# Comment only"], [], 0),
        (["- *"], [], 1),
        (["+fm:*/something00.txt", "-/data"], [], 2),
        (["R /"], ["/"], 0),
        (["R /", "# comment"], ["/"], 0),
        (["# comment", "- /data", "R /home"], ["/home"], 1),
    ],
)
def test_load_patterns_from_file(tmpdir, lines, expected_roots, expected_numpatterns):
    def evaluate(filename):
        roots = []
        inclexclpatterns = []
        load_pattern_file(open(filename), roots, inclexclpatterns)
        return roots, len(inclexclpatterns)

    patternfile = tmpdir.join("patterns.txt")

    with patternfile.open("wt") as fh:
        fh.write("\n".join(lines))

    roots, numpatterns = evaluate(str(patternfile))
    assert roots == expected_roots
    assert numpatterns == expected_numpatterns


def test_switch_patterns_style():
    patterns = """\
        +0_initial_default_is_shell
        p fm
        +1_fnmatch
        P re
        +2_regex
        +3_more_regex
        P pp
        +4_pathprefix
        p fm
        p sh
        +5_shell
    """
    pattern_file = io.StringIO(patterns)
    roots, patterns = [], []
    load_pattern_file(pattern_file, roots, patterns)
    assert len(patterns) == 6
    assert isinstance(patterns[0].val, ShellPattern)
    assert isinstance(patterns[1].val, FnmatchPattern)
    assert isinstance(patterns[2].val, RegexPattern)
    assert isinstance(patterns[3].val, RegexPattern)
    assert isinstance(patterns[4].val, PathPrefixPattern)
    assert isinstance(patterns[5].val, ShellPattern)


@pytest.mark.parametrize(
    "lines", [(["X /data"]), (["/data"])]  # illegal pattern type prefix  # need a pattern type prefix
)
def test_load_invalid_patterns_from_file(tmpdir, lines):
    patternfile = tmpdir.join("patterns.txt")
    with patternfile.open("wt") as fh:
        fh.write("\n".join(lines))
    filename = str(patternfile)
    with pytest.raises(argparse.ArgumentTypeError):
        roots = []
        inclexclpatterns = []
        load_pattern_file(open(filename), roots, inclexclpatterns)


@pytest.mark.parametrize(
    "lines, expected",
    [
        # "None" means all files, i.e. none excluded
        ([], None),
        (["# Comment only"], None),
        (["- *"], []),
        # default match type is sh: for patterns -> * doesn't match a /
        (
            ["-*/something0?.txt"],
            ["data", "data/subdir/something01.txt", "home", "home/leo", "home/leo/t", "home/other"],
        ),
        (
            ["-fm:*/something00.txt"],
            ["data", "data/subdir/something01.txt", "home", "home/leo", "home/leo/t", "home/other"],
        ),
        (["-fm:*/something0?.txt"], ["data", "home", "home/leo", "home/leo/t", "home/other"]),
        (["+/*/something0?.txt", "-/data"], ["data/something00.txt", "home", "home/leo", "home/leo/t", "home/other"]),
        (["+fm:*/something00.txt", "-/data"], ["data/something00.txt", "home", "home/leo", "home/leo/t", "home/other"]),
        # include /home/leo and exclude the rest of /home:
        (
            ["+/home/leo", "-/home/*"],
            ["data", "data/something00.txt", "data/subdir/something01.txt", "home", "home/leo", "home/leo/t"],
        ),
        # wrong order, /home/leo is already excluded by -/home/*:
        (["-/home/*", "+/home/leo"], ["data", "data/something00.txt", "data/subdir/something01.txt", "home"]),
        (
            ["+fm:/home/leo", "-/home/"],
            ["data", "data/something00.txt", "data/subdir/something01.txt", "home", "home/leo", "home/leo/t"],
        ),
    ],
)
def test_inclexcl_patterns_from_file(tmpdir, lines, expected):
    files = [
        "data",
        "data/something00.txt",
        "data/subdir/something01.txt",
        "home",
        "home/leo",
        "home/leo/t",
        "home/other",
    ]

    def evaluate(filename):
        matcher = PatternMatcher(fallback=True)
        roots = []
        inclexclpatterns = []
        load_pattern_file(open(filename), roots, inclexclpatterns)
        matcher.add_inclexcl(inclexclpatterns)
        return [path for path in files if matcher.match(path)]

    patternfile = tmpdir.join("patterns.txt")

    with patternfile.open("wt") as fh:
        fh.write("\n".join(lines))

    assert evaluate(str(patternfile)) == (files if expected is None else expected)


@pytest.mark.parametrize(
    "pattern, cls",
    [
        ("", FnmatchPattern),
        # Default style
        ("*", FnmatchPattern),
        ("/data/*", FnmatchPattern),
        # fnmatch style
        ("fm:", FnmatchPattern),
        ("fm:*", FnmatchPattern),
        ("fm:/data/*", FnmatchPattern),
        ("fm:fm:/data/*", FnmatchPattern),
        # Regular expression
        ("re:", RegexPattern),
        ("re:.*", RegexPattern),
        ("re:^/something/", RegexPattern),
        ("re:re:^/something/", RegexPattern),
        # Path prefix
        ("pp:", PathPrefixPattern),
        ("pp:/", PathPrefixPattern),
        ("pp:/data/", PathPrefixPattern),
        ("pp:pp:/data/", PathPrefixPattern),
        # Shell-pattern style
        ("sh:", ShellPattern),
        ("sh:*", ShellPattern),
        ("sh:/data/*", ShellPattern),
        ("sh:sh:/data/*", ShellPattern),
    ],
)
def test_parse_pattern(pattern, cls):
    assert isinstance(parse_pattern(pattern), cls)


@pytest.mark.parametrize("pattern", ["aa:", "fo:*", "00:", "x1:abc"])
def test_parse_pattern_error(pattern):
    with pytest.raises(ValueError):
        parse_pattern(pattern)


def test_pattern_matcher():
    pm = PatternMatcher()

    assert pm.fallback is None

    for i in ["", "foo", "bar"]:
        assert pm.match(i) is None

    # add extra entries to aid in testing
    for target in ["A", "B", "Empty", "FileNotFound"]:
        pm.is_include_cmd[target] = target

    pm.add([RegexPattern("^a")], "A")
    pm.add([RegexPattern("^b"), RegexPattern("^z")], "B")
    pm.add([RegexPattern("^$")], "Empty")
    pm.fallback = "FileNotFound"

    assert pm.match("") == "Empty"
    assert pm.match("aaa") == "A"
    assert pm.match("bbb") == "B"
    assert pm.match("ccc") == "FileNotFound"
    assert pm.match("xyz") == "FileNotFound"
    assert pm.match("z") == "B"

    assert PatternMatcher(fallback="hey!").fallback == "hey!"


@pytest.mark.parametrize(
    "pattern, regex",
    [
        ("foo.bar", r"foo\.bar"),  # default is id:
        ("id:foo.bar", r"foo\.bar"),
        ("id:foo?", r"foo\?"),
        ("re:foo.bar", r"foo.bar"),
        ("re:.*(fooo?|bar|baz).*", r".*(fooo?|bar|baz).*"),
        ("sh:foo.*", r"foo\.[^\/]*"),
    ],
)
def test_regex_from_pattern(pattern, regex):
    assert get_regex_from_pattern(pattern) == regex



import argparse

from ._common import with_repository, with_archive
from ..constants import *  # NOQA
from ..helpers import archivename_validator
from ..manifest import Manifest

from ..logger import create_logger

logger = create_logger()


class RenameMixIn:
    @with_repository(exclusive=True, cache=True, compatibility=(Manifest.Operation.CHECK,))
    @with_archive
    def do_rename(self, args, repository, manifest, cache, archive):
        """Rename an existing archive"""
        archive.rename(args.newname)
        manifest.write()
        repository.commit(compact=False)
        cache.commit()
        return self.exit_code

    def build_parser_rename(self, subparsers, common_parser, mid_common_parser):
        from ._common import process_epilog

        rename_epilog = process_epilog(
            """
        This command renames an archive in the repository.

        This results in a different archive ID.
        """
        )
        subparser = subparsers.add_parser(
            "rename",
            parents=[common_parser],
            add_help=False,
            description=self.do_rename.__doc__,
            epilog=rename_epilog,
            formatter_class=argparse.RawDescriptionHelpFormatter,
            help="rename archive",
        )
        subparser.set_defaults(func=self.do_rename)
        subparser.add_argument("name", metavar="OLDNAME", type=archivename_validator, help="specify the archive name")
        subparser.add_argument(
            "newname", metavar="NEWNAME", type=archivename_validator, help="specify the new archive name"
        )



import pytest

from ..crypto.key import PlaintextKey
from ..repository import Repository
from ..repoobj import RepoObj, RepoObj1
from ..compress import LZ4


@pytest.fixture
def repository(tmpdir):
    return Repository(tmpdir, create=True)


@pytest.fixture
def key(repository):
    return PlaintextKey(repository)


def test_format_parse_roundtrip(key):
    repo_objs = RepoObj(key)
    data = b"foobar" * 10
    id = repo_objs.id_hash(data)
    meta = {"custom": "something"}  # size and csize are computed automatically
    cdata = repo_objs.format(id, meta, data)

    got_meta = repo_objs.parse_meta(id, cdata)
    assert got_meta["size"] == len(data)
    assert got_meta["csize"] < len(data)
    assert got_meta["custom"] == "something"

    got_meta, got_data = repo_objs.parse(id, cdata)
    assert got_meta["size"] == len(data)
    assert got_meta["csize"] < len(data)
    assert got_meta["custom"] == "something"
    assert data == got_data

    edata = repo_objs.extract_crypted_data(cdata)
    key = repo_objs.key
    assert edata.startswith(bytes((key.TYPE,)))


def test_format_parse_roundtrip_borg1(key):  # legacy
    repo_objs = RepoObj1(key)
    data = b"foobar" * 10
    id = repo_objs.id_hash(data)
    meta = {}  # borg1 does not support this kind of metadata
    cdata = repo_objs.format(id, meta, data)

    # borg1 does not support separate metadata and borg2 does not invoke parse_meta for borg1 repos

    got_meta, got_data = repo_objs.parse(id, cdata)
    assert got_meta["size"] == len(data)
    assert got_meta["csize"] < len(data)
    assert data == got_data

    edata = repo_objs.extract_crypted_data(cdata)
    compressor = repo_objs.compressor
    key = repo_objs.key
    assert edata.startswith(bytes((key.TYPE, compressor.ID, compressor.level)))


def test_borg1_borg2_transition(key):
    # borg transfer reads borg 1.x repo objects (without decompressing them),
    # writes borg 2 repo objects (giving already compressed data to avoid compression).
    meta = {}  # borg1 does not support this kind of metadata
    data = b"foobar" * 10
    len_data = len(data)
    repo_objs1 = RepoObj1(key)
    id = repo_objs1.id_hash(data)
    borg1_cdata = repo_objs1.format(id, meta, data)
    meta1, compr_data1 = repo_objs1.parse(id, borg1_cdata, decompress=False)  # borg transfer avoids (de)compression
    # in borg 1, we can only get this metadata after decrypting the whole chunk (and we do not have "size" here):
    assert meta1["ctype"] == LZ4.ID  # default compression
    assert meta1["clevel"] == 0xFF  # lz4 does not know levels (yet?)
    assert meta1["csize"] < len_data  # lz4 should make it smaller

    repo_objs2 = RepoObj(key)
    # note: as we did not decompress, we do not have "size" and we need to get it from somewhere else.
    # here, we just use len_data. for borg transfer, we also know the size from another metadata source.
    borg2_cdata = repo_objs2.format(
        id, dict(meta1), compr_data1[2:], compress=False, size=len_data, ctype=meta1["ctype"], clevel=meta1["clevel"]
    )
    meta2, data2 = repo_objs2.parse(id, borg2_cdata)
    assert data2 == data
    assert meta2["ctype"] == LZ4.ID
    assert meta2["clevel"] == 0xFF
    assert meta2["csize"] == meta1["csize"] - 2  # borg2 does not store the type/level bytes there
    assert meta2["size"] == len_data

    meta2 = repo_objs2.parse_meta(id, borg2_cdata)
    # now, in borg 2, we have nice and separately decrypted metadata (no need to decrypt the whole chunk):
    assert meta2["ctype"] == LZ4.ID
    assert meta2["clevel"] == 0xFF
    assert meta2["csize"] == meta1["csize"] - 2  # borg2 does not store the type/level bytes there
    assert meta2["size"] == len_data



import logging
import shutil
import unittest
from unittest.mock import patch

from ...archive import ChunkBuffer
from ...constants import *  # NOQA
from ...helpers import bin_to_hex
from ...helpers import msgpack
from ...manifest import Manifest
from ...repository import Repository
from . import ArchiverTestCaseBase, RemoteArchiverTestCaseBase, ArchiverTestCaseBinaryBase, RK_ENCRYPTION, BORG_EXES


class ArchiverCheckTestCase(ArchiverTestCaseBase):
    def setUp(self):
        super().setUp()
        with patch.object(ChunkBuffer, "BUFFER_SIZE", 10):
            self.cmd(f"--repo={self.repository_location}", "rcreate", RK_ENCRYPTION)
            self.create_src_archive("archive1")
            self.create_src_archive("archive2")

    def test_check_usage(self):
        output = self.cmd(f"--repo={self.repository_location}", "check", "-v", "--progress", exit_code=0)
        self.assert_in("Starting repository check", output)
        self.assert_in("Starting archive consistency check", output)
        self.assert_in("Checking segments", output)
        # reset logging to new process default to avoid need for fork=True on next check
        logging.getLogger("borg.output.progress").setLevel(logging.NOTSET)
        output = self.cmd(f"--repo={self.repository_location}", "check", "-v", "--repository-only", exit_code=0)
        self.assert_in("Starting repository check", output)
        self.assert_not_in("Starting archive consistency check", output)
        self.assert_not_in("Checking segments", output)
        output = self.cmd(f"--repo={self.repository_location}", "check", "-v", "--archives-only", exit_code=0)
        self.assert_not_in("Starting repository check", output)
        self.assert_in("Starting archive consistency check", output)
        output = self.cmd(
            f"--repo={self.repository_location}",
            "check",
            "-v",
            "--archives-only",
            "--match-archives=archive2",
            exit_code=0,
        )
        self.assert_not_in("archive1", output)
        output = self.cmd(
            f"--repo={self.repository_location}", "check", "-v", "--archives-only", "--first=1", exit_code=0
        )
        self.assert_in("archive1", output)
        self.assert_not_in("archive2", output)
        output = self.cmd(
            f"--repo={self.repository_location}", "check", "-v", "--archives-only", "--last=1", exit_code=0
        )
        self.assert_not_in("archive1", output)
        self.assert_in("archive2", output)

    def test_date_matching(self):
        shutil.rmtree(self.repository_path)
        self.cmd(f"--repo={self.repository_location}", "rcreate", RK_ENCRYPTION)
        earliest_ts = "2022-11-20T23:59:59"
        ts_in_between = "2022-12-18T23:59:59"
        self.create_src_archive("archive1", ts=earliest_ts)
        self.create_src_archive("archive2", ts=ts_in_between)
        self.create_src_archive("archive3")
        output = self.cmd(
            f"--repo={self.repository_location}", "check", "-v", "--archives-only", "--oldest=23e", exit_code=2
        )
        output = self.cmd(
            f"--repo={self.repository_location}", "check", "-v", "--archives-only", "--oldest=1m", exit_code=0
        )
        self.assert_in("archive1", output)
        self.assert_in("archive2", output)
        self.assert_not_in("archive3", output)

        output = self.cmd(
            f"--repo={self.repository_location}", "check", "-v", "--archives-only", "--newest=1m", exit_code=0
        )
        self.assert_in("archive3", output)
        self.assert_not_in("archive2", output)
        self.assert_not_in("archive1", output)
        output = self.cmd(
            f"--repo={self.repository_location}", "check", "-v", "--archives-only", "--newer=1d", exit_code=0
        )
        self.assert_in("archive3", output)
        self.assert_not_in("archive1", output)
        self.assert_not_in("archive2", output)
        output = self.cmd(
            f"--repo={self.repository_location}", "check", "-v", "--archives-only", "--older=1d", exit_code=0
        )
        self.assert_in("archive1", output)
        self.assert_in("archive2", output)
        self.assert_not_in("archive3", output)

    def test_missing_file_chunk(self):
        archive, repository = self.open_archive("archive1")
        with repository:
            for item in archive.iter_items():
                if item.path.endswith("testsuite/archiver/__init__.py"):
                    valid_chunks = item.chunks
                    killed_chunk = valid_chunks[-1]
                    repository.delete(killed_chunk.id)
                    break
            else:
                self.fail("should not happen")
            repository.commit(compact=False)
        self.cmd(f"--repo={self.repository_location}", "check", exit_code=1)
        output = self.cmd(f"--repo={self.repository_location}", "check", "--repair", exit_code=0)
        self.assert_in("New missing file chunk detected", output)
        self.cmd(f"--repo={self.repository_location}", "check", exit_code=0)
        output = self.cmd(
            f"--repo={self.repository_location}", "list", "archive1", "--format={health}#{path}{NL}", exit_code=0
        )
        self.assert_in("broken#", output)
        # check that the file in the old archives has now a different chunk list without the killed chunk
        for archive_name in ("archive1", "archive2"):
            archive, repository = self.open_archive(archive_name)
            with repository:
                for item in archive.iter_items():
                    if item.path.endswith("testsuite/archiver/__init__.py"):
                        self.assert_not_equal(valid_chunks, item.chunks)
                        self.assert_not_in(killed_chunk, item.chunks)
                        break
                else:
                    self.fail("should not happen")
        # do a fresh backup (that will include the killed chunk)
        with patch.object(ChunkBuffer, "BUFFER_SIZE", 10):
            self.create_src_archive("archive3")
        # check should be able to heal the file now:
        output = self.cmd(f"--repo={self.repository_location}", "check", "-v", "--repair", exit_code=0)
        self.assert_in("Healed previously missing file chunk", output)
        self.assert_in("testsuite/archiver/__init__.py: Completely healed previously damaged file!", output)
        # check that the file in the old archives has the correct chunks again
        for archive_name in ("archive1", "archive2"):
            archive, repository = self.open_archive(archive_name)
            with repository:
                for item in archive.iter_items():
                    if item.path.endswith("testsuite/archiver/__init__.py"):
                        self.assert_equal(valid_chunks, item.chunks)
                        break
                else:
                    self.fail("should not happen")
        # list is also all-healthy again
        output = self.cmd(
            f"--repo={self.repository_location}", "list", "archive1", "--format={health}#{path}{NL}", exit_code=0
        )
        self.assert_not_in("broken#", output)

    def test_missing_archive_item_chunk(self):
        archive, repository = self.open_archive("archive1")
        with repository:
            repository.delete(archive.metadata.items[0])
            repository.commit(compact=False)
        self.cmd(f"--repo={self.repository_location}", "check", exit_code=1)
        self.cmd(f"--repo={self.repository_location}", "check", "--repair", exit_code=0)
        self.cmd(f"--repo={self.repository_location}", "check", exit_code=0)

    def test_missing_archive_metadata(self):
        archive, repository = self.open_archive("archive1")
        with repository:
            repository.delete(archive.id)
            repository.commit(compact=False)
        self.cmd(f"--repo={self.repository_location}", "check", exit_code=1)
        self.cmd(f"--repo={self.repository_location}", "check", "--repair", exit_code=0)
        self.cmd(f"--repo={self.repository_location}", "check", exit_code=0)

    def test_missing_manifest(self):
        archive, repository = self.open_archive("archive1")
        with repository:
            repository.delete(Manifest.MANIFEST_ID)
            repository.commit(compact=False)
        self.cmd(f"--repo={self.repository_location}", "check", exit_code=1)
        output = self.cmd(f"--repo={self.repository_location}", "check", "-v", "--repair", exit_code=0)
        self.assert_in("archive1", output)
        self.assert_in("archive2", output)
        self.cmd(f"--repo={self.repository_location}", "check", exit_code=0)

    def test_corrupted_manifest(self):
        archive, repository = self.open_archive("archive1")
        with repository:
            manifest = repository.get(Manifest.MANIFEST_ID)
            corrupted_manifest = manifest + b"corrupted!"
            repository.put(Manifest.MANIFEST_ID, corrupted_manifest)
            repository.commit(compact=False)
        self.cmd(f"--repo={self.repository_location}", "check", exit_code=1)
        output = self.cmd(f"--repo={self.repository_location}", "check", "-v", "--repair", exit_code=0)
        self.assert_in("archive1", output)
        self.assert_in("archive2", output)
        self.cmd(f"--repo={self.repository_location}", "check", exit_code=0)

    def test_manifest_rebuild_corrupted_chunk(self):
        archive, repository = self.open_archive("archive1")
        with repository:
            manifest = repository.get(Manifest.MANIFEST_ID)
            corrupted_manifest = manifest + b"corrupted!"
            repository.put(Manifest.MANIFEST_ID, corrupted_manifest)

            chunk = repository.get(archive.id)
            corrupted_chunk = chunk + b"corrupted!"
            repository.put(archive.id, corrupted_chunk)
            repository.commit(compact=False)
        self.cmd(f"--repo={self.repository_location}", "check", exit_code=1)
        output = self.cmd(f"--repo={self.repository_location}", "check", "-v", "--repair", exit_code=0)
        self.assert_in("archive2", output)
        self.cmd(f"--repo={self.repository_location}", "check", exit_code=0)

    def test_manifest_rebuild_duplicate_archive(self):
        archive, repository = self.open_archive("archive1")
        repo_objs = archive.repo_objs

        with repository:
            manifest = repository.get(Manifest.MANIFEST_ID)
            corrupted_manifest = manifest + b"corrupted!"
            repository.put(Manifest.MANIFEST_ID, corrupted_manifest)

            archive = msgpack.packb(
                {
                    "command_line": "",
                    "item_ptrs": [],
                    "hostname": "foo",
                    "username": "bar",
                    "name": "archive1",
                    "time": "2016-12-15T18:49:51.849711",
                    "version": 2,
                }
            )
            archive_id = repo_objs.id_hash(archive)
            repository.put(archive_id, repo_objs.format(archive_id, {}, archive))
            repository.commit(compact=False)
        self.cmd(f"--repo={self.repository_location}", "check", exit_code=1)
        self.cmd(f"--repo={self.repository_location}", "check", "--repair", exit_code=0)
        output = self.cmd(f"--repo={self.repository_location}", "rlist")
        self.assert_in("archive1", output)
        self.assert_in("archive1.1", output)
        self.assert_in("archive2", output)

    def test_extra_chunks(self):
        self.cmd(f"--repo={self.repository_location}", "check", exit_code=0)
        with Repository(self.repository_location, exclusive=True) as repository:
            repository.put(b"01234567890123456789012345678901", b"xxxx")
            repository.commit(compact=False)
        self.cmd(f"--repo={self.repository_location}", "check", exit_code=1)
        self.cmd(f"--repo={self.repository_location}", "check", exit_code=1)
        self.cmd(f"--repo={self.repository_location}", "check", "--repair", exit_code=0)
        self.cmd(f"--repo={self.repository_location}", "check", exit_code=0)
        self.cmd(f"--repo={self.repository_location}", "extract", "archive1", "--dry-run", exit_code=0)

    def _test_verify_data(self, *init_args):
        shutil.rmtree(self.repository_path)
        self.cmd(f"--repo={self.repository_location}", "rcreate", *init_args)
        self.create_src_archive("archive1")
        archive, repository = self.open_archive("archive1")
        with repository:
            for item in archive.iter_items():
                if item.path.endswith("testsuite/archiver/__init__.py"):
                    chunk = item.chunks[-1]
                    data = repository.get(chunk.id)
                    data = data[0:100] + b"x" + data[101:]
                    repository.put(chunk.id, data)
                    break
            repository.commit(compact=False)
        self.cmd(f"--repo={self.repository_location}", "check", exit_code=0)
        output = self.cmd(f"--repo={self.repository_location}", "check", "--verify-data", exit_code=1)
        assert bin_to_hex(chunk.id) + ", integrity error" in output
        # repair (heal is tested in another test)
        output = self.cmd(f"--repo={self.repository_location}", "check", "--repair", "--verify-data", exit_code=0)
        assert bin_to_hex(chunk.id) + ", integrity error" in output
        assert "testsuite/archiver/__init__.py: New missing file chunk detected" in output

    def test_verify_data(self):
        self._test_verify_data(RK_ENCRYPTION)

    def test_verify_data_unencrypted(self):
        self._test_verify_data("--encryption", "none")

    def test_empty_repository(self):
        with Repository(self.repository_location, exclusive=True) as repository:
            for id_ in repository.list():
                repository.delete(id_)
            repository.commit(compact=False)
        self.cmd(f"--repo={self.repository_location}", "check", exit_code=1)


class RemoteArchiverCheckTestCase(RemoteArchiverTestCaseBase, ArchiverCheckTestCase):
    """run the same tests, but with a remote repository"""

    @unittest.skip("only works locally")
    def test_empty_repository(self):
        pass

    @unittest.skip("only works locally")
    def test_extra_chunks(self):
        pass


@unittest.skipUnless("binary" in BORG_EXES, "no borg.exe available")
class ArchiverTestCaseBinary(ArchiverTestCaseBinaryBase, ArchiverCheckTestCase):
    """runs the same tests, but via the borg binary"""



import os
import zlib

try:
    import lzma
except ImportError:
    lzma = None

import pytest

from ..compress import get_compressor, Compressor, CompressionSpec, CNONE, ZLIB, LZ4, LZMA, ZSTD, Auto


buffer = bytes(2**16)
data = b"fooooooooobaaaaaaaar" * 10
params = dict(name="zlib", level=6)


def test_get_compressor():
    c = get_compressor(name="none")
    assert isinstance(c, CNONE)
    c = get_compressor(name="lz4")
    assert isinstance(c, LZ4)
    c = get_compressor(name="zlib")
    assert isinstance(c, ZLIB)
    with pytest.raises(KeyError):
        get_compressor(name="foobar")


def test_cnull():
    c = get_compressor(name="none")
    meta, cdata = c.compress({}, data)
    assert len(cdata) >= len(data)
    assert data in cdata  # it's not compressed and just in there 1:1
    assert data == c.decompress(meta, cdata)[1]
    assert data == Compressor(**params).decompress(meta, cdata)[1]  # autodetect


def test_lz4():
    c = get_compressor(name="lz4")
    meta, cdata = c.compress({}, data)
    assert len(cdata) < len(data)
    assert data == c.decompress(meta, cdata)[1]
    assert data == Compressor(**params).decompress(meta, cdata)[1]  # autodetect


def test_lz4_buffer_allocation(monkeypatch):
    # disable fallback to no compression on incompressible data
    monkeypatch.setattr(LZ4, "decide", lambda always_compress: LZ4)
    # test with a rather huge data object to see if buffer allocation / resizing works
    data = os.urandom(5 * 2**20) * 10  # 50MiB badly compressible data
    assert len(data) == 50 * 2**20
    c = Compressor("lz4")
    meta, cdata = c.compress({}, data)
    assert len(cdata) >= len(data)
    assert data == c.decompress(meta, cdata)[1]


def test_zlib():
    c = get_compressor(name="zlib")
    meta, cdata = c.compress({}, data)
    assert len(cdata) < len(data)
    assert data == c.decompress(meta, cdata)[1]
    assert data == Compressor(**params).decompress(meta, cdata)[1]  # autodetect


def test_lzma():
    if lzma is None:
        pytest.skip("No lzma support found.")
    c = get_compressor(name="lzma")
    meta, cdata = c.compress({}, data)
    assert len(cdata) < len(data)
    assert data == c.decompress(meta, cdata)[1]
    assert data == Compressor(**params).decompress(meta, cdata)[1]  # autodetect


def test_zstd():
    c = get_compressor(name="zstd")
    meta, cdata = c.compress({}, data)
    assert len(cdata) < len(data)
    assert data == c.decompress(meta, cdata)[1]
    assert data == Compressor(**params).decompress(meta, cdata)[1]  # autodetect


def test_autodetect_invalid():
    with pytest.raises(ValueError):
        Compressor(**params, legacy_mode=True).decompress({}, b"\xff\xfftotalcrap")
    with pytest.raises(ValueError):
        Compressor(**params, legacy_mode=True).decompress({}, b"\x08\x00notreallyzlib")


def test_zlib_legacy_compat():
    # for compatibility reasons, we do not add an extra header for zlib,
    # nor do we expect one when decompressing / autodetecting
    for level in range(10):
        c = get_compressor(name="zlib_legacy", level=level, legacy_mode=True)
        meta1, cdata1 = c.compress({}, data)
        cdata2 = zlib.compress(data, level)
        assert cdata1 == cdata2
        meta2, data2 = c.decompress({}, cdata2)
        assert data == data2
        # _, data2 = Compressor(**params).decompress({}, cdata2)
        # assert data == data2


def test_compressor():
    params_list = [
        dict(name="none"),
        dict(name="lz4"),
        dict(name="zstd", level=1),
        dict(name="zstd", level=3),
        # avoiding high zstd levels, memory needs unclear
        dict(name="zlib", level=0),
        dict(name="zlib", level=6),
        dict(name="zlib", level=9),
    ]
    if lzma:
        params_list += [
            dict(name="lzma", level=0),
            dict(name="lzma", level=6),
            # we do not test lzma on level 9 because of the huge memory needs
        ]
    for params in params_list:
        c = Compressor(**params)
        meta_c, data_compressed = c.compress({}, data)
        assert "ctype" in meta_c
        assert "clevel" in meta_c
        assert meta_c["csize"] == len(data_compressed)
        assert meta_c["size"] == len(data)
        meta_d, data_decompressed = c.decompress(meta_c, data_compressed)
        assert data == data_decompressed
        assert "ctype" in meta_d
        assert "clevel" in meta_d
        assert meta_d["csize"] == len(data_compressed)
        assert meta_d["size"] == len(data)


def test_auto():
    compressor_auto_zlib = CompressionSpec("auto,zlib,9").compressor
    compressor_lz4 = CompressionSpec("lz4").compressor
    compressor_zlib = CompressionSpec("zlib,9").compressor
    data = bytes(500)
    meta, compressed_auto_zlib = compressor_auto_zlib.compress({}, data)
    _, compressed_lz4 = compressor_lz4.compress({}, data)
    _, compressed_zlib = compressor_zlib.compress({}, data)
    ratio = len(compressed_zlib) / len(compressed_lz4)
    assert meta["ctype"] == ZLIB.ID if ratio < 0.99 else LZ4.ID
    assert meta["clevel"] == 9 if ratio < 0.99 else 255
    assert meta["csize"] == len(compressed_auto_zlib)

    data = b"\x00\xb8\xa3\xa2-O\xe1i\xb6\x12\x03\xc21\xf3\x8a\xf78\\\x01\xa5b\x07\x95\xbeE\xf8\xa3\x9ahm\xb1~"
    meta, compressed = compressor_auto_zlib.compress(dict(meta), data)
    assert meta["ctype"] == CNONE.ID
    assert meta["clevel"] == 255
    assert meta["csize"] == len(compressed)


def test_obfuscate():
    compressor = CompressionSpec("obfuscate,1,none").compressor
    data = bytes(10000)
    _, compressed = compressor.compress({}, data)
    assert len(data) <= len(compressed) <= len(data) * 101
    # compressing 100 times the same data should give at least 50 different result sizes
    assert len({len(compressor.compress({}, data)[1]) for i in range(100)}) > 50

    cs = CompressionSpec("obfuscate,2,lz4")
    assert isinstance(cs.inner.compressor, LZ4)
    compressor = cs.compressor
    data = bytes(10000)
    _, compressed = compressor.compress({}, data)
    min_compress, max_compress = 0.2, 0.001  # estimate compression factor outer boundaries
    assert max_compress * len(data) <= len(compressed) <= min_compress * len(data) * 1001
    # compressing 100 times the same data should give multiple different result sizes
    assert len({len(compressor.compress({}, data)[1]) for i in range(100)}) > 10

    cs = CompressionSpec("obfuscate,6,zstd,3")
    assert isinstance(cs.inner.compressor, ZSTD)
    compressor = cs.compressor
    data = bytes(10000)
    _, compressed = compressor.compress({}, data)
    min_compress, max_compress = 0.2, 0.001  # estimate compression factor outer boundaries
    assert max_compress * len(data) <= len(compressed) <= min_compress * len(data) * 10000001
    # compressing 100 times the same data should give multiple different result sizes
    assert len({len(compressor.compress({}, data)[1]) for i in range(100)}) > 90

    cs = CompressionSpec("obfuscate,2,auto,zstd,10")
    assert isinstance(cs.inner.compressor, Auto)
    compressor = cs.compressor
    data = bytes(10000)
    _, compressed = compressor.compress({}, data)
    min_compress, max_compress = 0.2, 0.001  # estimate compression factor outer boundaries
    assert max_compress * len(data) <= len(compressed) <= min_compress * len(data) * 1001
    # compressing 100 times the same data should give multiple different result sizes
    assert len({len(compressor.compress({}, data)[1]) for i in range(100)}) > 10

    cs = CompressionSpec("obfuscate,110,none")
    assert isinstance(cs.inner.compressor, CNONE)
    compressor = cs.compressor
    data = bytes(1000)
    _, compressed = compressor.compress({}, data)
    assert 1000 <= len(compressed) <= 1000 + 1024
    data = bytes(1100)
    _, compressed = compressor.compress({}, data)
    assert 1100 <= len(compressed) <= 1100 + 1024


def test_obfuscate_meta():
    compressor = CompressionSpec("obfuscate,3,lz4").compressor
    meta = {}
    data = bytes(10000)
    meta, compressed = compressor.compress(meta, data)
    assert "ctype" in meta
    assert meta["ctype"] == LZ4.ID
    assert "clevel" in meta
    assert meta["clevel"] == 0xFF
    assert "csize" in meta
    csize = meta["csize"]
    assert csize == len(compressed)  # this is the overall size
    assert "psize" in meta
    psize = meta["psize"]
    assert 0 < psize < 100
    assert csize - psize >= 0  # there is a obfuscation trailer
    trailer = compressed[psize:]
    assert not trailer or set(trailer) == {0}  # trailer is all-zero-bytes


def test_compression_specs():
    with pytest.raises(ValueError):
        CompressionSpec("")

    assert isinstance(CompressionSpec("none").compressor, CNONE)
    assert isinstance(CompressionSpec("lz4").compressor, LZ4)

    zlib = CompressionSpec("zlib").compressor
    assert isinstance(zlib, ZLIB)
    assert zlib.level == 6
    zlib = CompressionSpec("zlib,0").compressor
    assert isinstance(zlib, ZLIB)
    assert zlib.level == 0
    zlib = CompressionSpec("zlib,9").compressor
    assert isinstance(zlib, ZLIB)
    assert zlib.level == 9
    with pytest.raises(ValueError):
        CompressionSpec("zlib,9,invalid")

    lzma = CompressionSpec("lzma").compressor
    assert isinstance(lzma, LZMA)
    assert lzma.level == 6
    lzma = CompressionSpec("lzma,0").compressor
    assert isinstance(lzma, LZMA)
    assert lzma.level == 0
    lzma = CompressionSpec("lzma,9").compressor
    assert isinstance(lzma, LZMA)
    assert lzma.level == 9

    zstd = CompressionSpec("zstd").compressor
    assert isinstance(zstd, ZSTD)
    assert zstd.level == 3
    zstd = CompressionSpec("zstd,1").compressor
    assert isinstance(zstd, ZSTD)
    assert zstd.level == 1
    zstd = CompressionSpec("zstd,22").compressor
    assert isinstance(zstd, ZSTD)
    assert zstd.level == 22

    with pytest.raises(ValueError):
        CompressionSpec("lzma,9,invalid")
    with pytest.raises(ValueError):
        CompressionSpec("invalid")



import argparse
from contextlib import contextmanager
import functools
import os
import shutil
import time

from ..constants import *  # NOQA
from ..crypto.key import FlexiKey
from ..helpers import format_file_size
from ..helpers import msgpack
from ..item import Item
from ..platform import SyncFile


class BenchmarkMixIn:
    def do_benchmark_crud(self, args):
        """Benchmark Create, Read, Update, Delete for archives."""

        def measurement_run(repo, path):
            compression = "--compression=none"
            # measure create perf (without files cache to always have it chunking)
            t_start = time.monotonic()
            rc = self.do_create(
                self.parse_args(
                    [f"--repo={repo}", "create", compression, "--files-cache=disabled", "borg-benchmark-crud1", path]
                )
            )
            t_end = time.monotonic()
            dt_create = t_end - t_start
            assert rc == 0
            # now build files cache
            rc1 = self.do_create(
                self.parse_args([f"--repo={repo}", "create", compression, "borg-benchmark-crud2", path])
            )
            rc2 = self.do_delete(self.parse_args([f"--repo={repo}", "delete", "-a", "borg-benchmark-crud2"]))
            assert rc1 == rc2 == 0
            # measure a no-change update (archive1 is still present)
            t_start = time.monotonic()
            rc1 = self.do_create(
                self.parse_args([f"--repo={repo}", "create", compression, "borg-benchmark-crud3", path])
            )
            t_end = time.monotonic()
            dt_update = t_end - t_start
            rc2 = self.do_delete(self.parse_args([f"--repo={repo}", "delete", "-a", "borg-benchmark-crud3"]))
            assert rc1 == rc2 == 0
            # measure extraction (dry-run: without writing result to disk)
            t_start = time.monotonic()
            rc = self.do_extract(self.parse_args([f"--repo={repo}", "extract", "borg-benchmark-crud1", "--dry-run"]))
            t_end = time.monotonic()
            dt_extract = t_end - t_start
            assert rc == 0
            # measure archive deletion (of LAST present archive with the data)
            t_start = time.monotonic()
            rc = self.do_delete(self.parse_args([f"--repo={repo}", "delete", "-a", "borg-benchmark-crud1"]))
            t_end = time.monotonic()
            dt_delete = t_end - t_start
            assert rc == 0
            return dt_create, dt_update, dt_extract, dt_delete

        @contextmanager
        def test_files(path, count, size, random):
            try:
                path = os.path.join(path, "borg-test-data")
                os.makedirs(path)
                z_buff = None if random else memoryview(zeros)[:size] if size <= len(zeros) else b"\0" * size
                for i in range(count):
                    fname = os.path.join(path, "file_%d" % i)
                    data = z_buff if not random else os.urandom(size)
                    with SyncFile(fname, binary=True) as fd:  # used for posix_fadvise's sake
                        fd.write(data)
                yield path
            finally:
                shutil.rmtree(path)

        if "_BORG_BENCHMARK_CRUD_TEST" in os.environ:
            tests = [("Z-TEST", 1, 1, False), ("R-TEST", 1, 1, True)]
        else:
            tests = [
                ("Z-BIG", 10, 100000000, False),
                ("R-BIG", 10, 100000000, True),
                ("Z-MEDIUM", 1000, 1000000, False),
                ("R-MEDIUM", 1000, 1000000, True),
                ("Z-SMALL", 10000, 10000, False),
                ("R-SMALL", 10000, 10000, True),
            ]

        for msg, count, size, random in tests:
            with test_files(args.path, count, size, random) as path:
                dt_create, dt_update, dt_extract, dt_delete = measurement_run(args.location.canonical_path(), path)
            total_size_MB = count * size / 1e06
            file_size_formatted = format_file_size(size)
            content = "random" if random else "all-zero"
            fmt = "%s-%-10s %9.2f MB/s (%d * %s %s files: %.2fs)"
            print(fmt % ("C", msg, total_size_MB / dt_create, count, file_size_formatted, content, dt_create))
            print(fmt % ("R", msg, total_size_MB / dt_extract, count, file_size_formatted, content, dt_extract))
            print(fmt % ("U", msg, total_size_MB / dt_update, count, file_size_formatted, content, dt_update))
            print(fmt % ("D", msg, total_size_MB / dt_delete, count, file_size_formatted, content, dt_delete))

        return 0

    def do_benchmark_cpu(self, args):
        """Benchmark CPU bound operations."""
        from timeit import timeit

        random_10M = os.urandom(10 * 1000 * 1000)
        key_256 = os.urandom(32)
        key_128 = os.urandom(16)
        key_96 = os.urandom(12)

        import io
        from ..chunker import get_chunker

        print("Chunkers =======================================================")
        size = "1GB"

        def chunkit(chunker_name, *args, **kwargs):
            with io.BytesIO(random_10M) as data_file:
                ch = get_chunker(chunker_name, *args, **kwargs)
                for _ in ch.chunkify(fd=data_file):
                    pass

        for spec, func in [
            ("buzhash,19,23,21,4095", lambda: chunkit("buzhash", 19, 23, 21, 4095, seed=0)),
            ("fixed,1048576", lambda: chunkit("fixed", 1048576, sparse=False)),
        ]:
            print(f"{spec:<24} {size:<10} {timeit(func, number=100):.3f}s")

        from ..checksums import crc32, xxh64

        print("Non-cryptographic checksums / hashes ===========================")
        size = "1GB"
        tests = [("xxh64", lambda: xxh64(random_10M)), ("crc32 (zlib)", lambda: crc32(random_10M))]
        for spec, func in tests:
            print(f"{spec:<24} {size:<10} {timeit(func, number=100):.3f}s")

        from ..crypto.low_level import hmac_sha256, blake2b_256

        print("Cryptographic hashes / MACs ====================================")
        size = "1GB"
        for spec, func in [
            ("hmac-sha256", lambda: hmac_sha256(key_256, random_10M)),
            ("blake2b-256", lambda: blake2b_256(key_256, random_10M)),
        ]:
            print(f"{spec:<24} {size:<10} {timeit(func, number=100):.3f}s")

        from ..crypto.low_level import AES256_CTR_BLAKE2b, AES256_CTR_HMAC_SHA256
        from ..crypto.low_level import AES256_OCB, CHACHA20_POLY1305

        print("Encryption =====================================================")
        size = "1GB"

        tests = [
            (
                "aes-256-ctr-hmac-sha256",
                lambda: AES256_CTR_HMAC_SHA256(key_256, key_256, iv=key_128, header_len=1, aad_offset=1).encrypt(
                    random_10M, header=b"X"
                ),
            ),
            (
                "aes-256-ctr-blake2b",
                lambda: AES256_CTR_BLAKE2b(key_256 * 4, key_256, iv=key_128, header_len=1, aad_offset=1).encrypt(
                    random_10M, header=b"X"
                ),
            ),
            (
                "aes-256-ocb",
                lambda: AES256_OCB(key_256, iv=key_96, header_len=1, aad_offset=1).encrypt(random_10M, header=b"X"),
            ),
            (
                "chacha20-poly1305",
                lambda: CHACHA20_POLY1305(key_256, iv=key_96, header_len=1, aad_offset=1).encrypt(
                    random_10M, header=b"X"
                ),
            ),
        ]
        for spec, func in tests:
            print(f"{spec:<24} {size:<10} {timeit(func, number=100):.3f}s")

        print("KDFs (slow is GOOD, use argon2!) ===============================")
        count = 5
        for spec, func in [
            ("pbkdf2", lambda: FlexiKey.pbkdf2("mypassphrase", b"salt" * 8, PBKDF2_ITERATIONS, 32)),
            ("argon2", lambda: FlexiKey.argon2("mypassphrase", 64, b"S" * ARGON2_SALT_BYTES, **ARGON2_ARGS)),
        ]:
            print(f"{spec:<24} {count:<10} {timeit(func, number=count):.3f}s")

        from ..compress import CompressionSpec

        print("Compression ====================================================")
        for spec in [
            "lz4",
            "zstd,1",
            "zstd,3",
            "zstd,5",
            "zstd,10",
            "zstd,16",
            "zstd,22",
            "zlib,0",
            "zlib,6",
            "zlib,9",
            "lzma,0",
            "lzma,6",
            "lzma,9",
        ]:
            compressor = CompressionSpec(spec).compressor
            size = "0.1GB"
            print(f"{spec:<12} {size:<10} {timeit(lambda: compressor.compress({}, random_10M), number=10):.3f}s")

        print("msgpack ========================================================")
        item = Item(path="/foo/bar/baz", mode=660, mtime=1234567)
        items = [item.as_dict()] * 1000
        size = "100k Items"
        spec = "msgpack"
        print(f"{spec:<12} {size:<10} {timeit(lambda: msgpack.packb(items), number=100):.3f}s")

        return 0

    def build_parser_benchmarks(self, subparsers, common_parser, mid_common_parser):
        from ._common import process_epilog

        benchmark_epilog = process_epilog("These commands do various benchmarks.")

        subparser = subparsers.add_parser(
            "benchmark",
            parents=[mid_common_parser],
            add_help=False,
            description="benchmark command",
            epilog=benchmark_epilog,
            formatter_class=argparse.RawDescriptionHelpFormatter,
            help="benchmark command",
        )

        benchmark_parsers = subparser.add_subparsers(title="required arguments", metavar="<command>")
        subparser.set_defaults(fallback_func=functools.partial(self.do_subcommand_help, subparser))

        bench_crud_epilog = process_epilog(
            """
        This command benchmarks borg CRUD (create, read, update, delete) operations.

        It creates input data below the given PATH and backups this data into the given REPO.
        The REPO must already exist (it could be a fresh empty repo or an existing repo, the
        command will create / read / update / delete some archives named borg-benchmark-crud\\* there.

        Make sure you have free space there, you'll need about 1GB each (+ overhead).

        If your repository is encrypted and borg needs a passphrase to unlock the key, use::

            BORG_PASSPHRASE=mysecret borg benchmark crud REPO PATH

        Measurements are done with different input file sizes and counts.
        The file contents are very artificial (either all zero or all random),
        thus the measurement results do not necessarily reflect performance with real data.
        Also, due to the kind of content used, no compression is used in these benchmarks.

        C- == borg create (1st archive creation, no compression, do not use files cache)
              C-Z- == all-zero files. full dedup, this is primarily measuring reader/chunker/hasher.
              C-R- == random files. no dedup, measuring throughput through all processing stages.

        R- == borg extract (extract archive, dry-run, do everything, but do not write files to disk)
              R-Z- == all zero files. Measuring heavily duplicated files.
              R-R- == random files. No duplication here, measuring throughput through all processing
              stages, except writing to disk.

        U- == borg create (2nd archive creation of unchanged input files, measure files cache speed)
              The throughput value is kind of virtual here, it does not actually read the file.
              U-Z- == needs to check the 2 all-zero chunks' existence in the repo.
              U-R- == needs to check existence of a lot of different chunks in the repo.

        D- == borg delete archive (delete last remaining archive, measure deletion + compaction)
              D-Z- == few chunks to delete / few segments to compact/remove.
              D-R- == many chunks to delete / many segments to compact/remove.

        Please note that there might be quite some variance in these measurements.
        Try multiple measurements and having a otherwise idle machine (and network, if you use it).
        """
        )
        subparser = benchmark_parsers.add_parser(
            "crud",
            parents=[common_parser],
            add_help=False,
            description=self.do_benchmark_crud.__doc__,
            epilog=bench_crud_epilog,
            formatter_class=argparse.RawDescriptionHelpFormatter,
            help="benchmarks borg CRUD (create, extract, update, delete).",
        )
        subparser.set_defaults(func=self.do_benchmark_crud)

        subparser.add_argument("path", metavar="PATH", help="path were to create benchmark input data")

        bench_cpu_epilog = process_epilog(
            """
        This command benchmarks misc. CPU bound borg operations.

        It creates input data in memory, runs the operation and then displays throughput.
        To reduce outside influence on the timings, please make sure to run this with:

        - an otherwise as idle as possible machine
        - enough free memory so there will be no slow down due to paging activity
        """
        )
        subparser = benchmark_parsers.add_parser(
            "cpu",
            parents=[common_parser],
            add_help=False,
            description=self.do_benchmark_cpu.__doc__,
            epilog=bench_cpu_epilog,
            formatter_class=argparse.RawDescriptionHelpFormatter,
            help="benchmarks borg CPU bound operations.",
        )
        subparser.set_defaults(func=self.do_benchmark_cpu)



import json
import os
import unittest

from ...constants import *  # NOQA
from . import (
    ArchiverTestCaseBase,
    RemoteArchiverTestCaseBase,
    ArchiverTestCaseBinaryBase,
    src_dir,
    RK_ENCRYPTION,
    checkts,
    BORG_EXES,
)


class ArchiverTestCase(ArchiverTestCaseBase):
    def test_rlist_glob(self):
        self.cmd(f"--repo={self.repository_location}", "rcreate", RK_ENCRYPTION)
        self.cmd(f"--repo={self.repository_location}", "create", "test-1", src_dir)
        self.cmd(f"--repo={self.repository_location}", "create", "something-else-than-test-1", src_dir)
        self.cmd(f"--repo={self.repository_location}", "create", "test-2", src_dir)
        output = self.cmd(f"--repo={self.repository_location}", "rlist", "--match-archives=sh:test-*")
        self.assert_in("test-1", output)
        self.assert_in("test-2", output)
        self.assert_not_in("something-else", output)

    def test_archives_format(self):
        self.cmd(f"--repo={self.repository_location}", "rcreate", RK_ENCRYPTION)
        self.cmd(f"--repo={self.repository_location}", "create", "--comment", "comment 1", "test-1", src_dir)
        self.cmd(f"--repo={self.repository_location}", "create", "--comment", "comment 2", "test-2", src_dir)
        output_1 = self.cmd(f"--repo={self.repository_location}", "rlist")
        output_2 = self.cmd(
            f"--repo={self.repository_location}", "rlist", "--format", "{archive:<36} {time} [{id}]{NL}"
        )
        self.assertEqual(output_1, output_2)
        output_1 = self.cmd(f"--repo={self.repository_location}", "rlist", "--short")
        self.assertEqual(output_1, "test-1" + os.linesep + "test-2" + os.linesep)
        output_3 = self.cmd(f"--repo={self.repository_location}", "rlist", "--format", "{name} {comment}{NL}")
        self.assert_in("test-1 comment 1" + os.linesep, output_3)
        self.assert_in("test-2 comment 2" + os.linesep, output_3)

    def test_date_matching(self):
        self.cmd(f"--repo={self.repository_location}", "rcreate", RK_ENCRYPTION)
        earliest_ts = "2022-11-20T23:59:59"
        ts_in_between = "2022-12-18T23:59:59"
        self.create_src_archive("archive1", ts=earliest_ts)
        self.create_src_archive("archive2", ts=ts_in_between)
        self.create_src_archive("archive3")
        output = self.cmd(f"--repo={self.repository_location}", "rlist", "-v", "--oldest=23e", exit_code=2)
        output = self.cmd(f"--repo={self.repository_location}", "rlist", "-v", "--oldest=1m", exit_code=0)
        self.assert_in("archive1", output)
        self.assert_in("archive2", output)
        self.assert_not_in("archive3", output)

        output = self.cmd(f"--repo={self.repository_location}", "rlist", "-v", "--newest=1m", exit_code=0)
        self.assert_in("archive3", output)
        self.assert_not_in("archive2", output)
        self.assert_not_in("archive1", output)
        output = self.cmd(f"--repo={self.repository_location}", "rlist", "-v", "--newer=1d", exit_code=0)
        self.assert_in("archive3", output)
        self.assert_not_in("archive1", output)
        self.assert_not_in("archive2", output)
        output = self.cmd(f"--repo={self.repository_location}", "rlist", "-v", "--older=1d", exit_code=0)
        self.assert_in("archive1", output)
        self.assert_in("archive2", output)
        self.assert_not_in("archive3", output)

    def test_rlist_consider_checkpoints(self):
        self.cmd(f"--repo={self.repository_location}", "rcreate", RK_ENCRYPTION)
        self.cmd(f"--repo={self.repository_location}", "create", "test1", src_dir)
        # these are not really a checkpoints, but they look like some:
        self.cmd(f"--repo={self.repository_location}", "create", "test2.checkpoint", src_dir)
        self.cmd(f"--repo={self.repository_location}", "create", "test3.checkpoint.1", src_dir)
        output = self.cmd(f"--repo={self.repository_location}", "rlist")
        assert "test1" in output
        assert "test2.checkpoint" not in output
        assert "test3.checkpoint.1" not in output
        output = self.cmd(f"--repo={self.repository_location}", "rlist", "--consider-checkpoints")
        assert "test1" in output
        assert "test2.checkpoint" in output
        assert "test3.checkpoint.1" in output

    def test_rlist_json(self):
        self.create_regular_file("file1", size=1024 * 80)
        self.cmd(f"--repo={self.repository_location}", "rcreate", RK_ENCRYPTION)
        self.cmd(f"--repo={self.repository_location}", "create", "test", "input")

        list_repo = json.loads(self.cmd(f"--repo={self.repository_location}", "rlist", "--json"))
        repository = list_repo["repository"]
        assert len(repository["id"]) == 64
        checkts(repository["last_modified"])
        assert list_repo["encryption"]["mode"] == RK_ENCRYPTION[13:]
        assert "keyfile" not in list_repo["encryption"]
        archive0 = list_repo["archives"][0]
        checkts(archive0["time"])


class RemoteArchiverTestCase(RemoteArchiverTestCaseBase, ArchiverTestCase):
    """run the same tests, but with a remote repository"""


@unittest.skipUnless("binary" in BORG_EXES, "no borg.exe available")
class ArchiverTestCaseBinary(ArchiverTestCaseBinaryBase, ArchiverTestCase):
    """runs the same tests, but via the borg binary"""



import os
import shutil
import unittest
from datetime import datetime, timezone, timedelta
from unittest.mock import patch

import pytest

from ...cache import Cache, LocalCache
from ...constants import *  # NOQA
from ...crypto.key import TAMRequiredError
from ...helpers import Location, get_security_dir
from ...helpers import EXIT_ERROR
from ...helpers import bin_to_hex
from ...helpers import msgpack
from ...manifest import Manifest, MandatoryFeatureUnsupported
from ...remote import RemoteRepository, PathNotAllowed
from ...repository import Repository
from .. import llfuse
from .. import changedir, environment_variable
from . import ArchiverTestCaseBase, RemoteArchiverTestCaseBase, RK_ENCRYPTION


class ArchiverTestCase(ArchiverTestCaseBase):
    def get_security_dir(self):
        repository_id = bin_to_hex(self._extract_repository_id(self.repository_path))
        return get_security_dir(repository_id)

    def test_repository_swap_detection(self):
        self.create_test_files()
        os.environ["BORG_PASSPHRASE"] = "passphrase"
        self.cmd(f"--repo={self.repository_location}", "rcreate", RK_ENCRYPTION)
        repository_id = self._extract_repository_id(self.repository_path)
        self.cmd(f"--repo={self.repository_location}", "create", "test", "input")
        shutil.rmtree(self.repository_path)
        self.cmd(f"--repo={self.repository_location}", "rcreate", "--encryption=none")
        self._set_repository_id(self.repository_path, repository_id)
        self.assert_equal(repository_id, self._extract_repository_id(self.repository_path))
        if self.FORK_DEFAULT:
            self.cmd(f"--repo={self.repository_location}", "create", "test.2", "input", exit_code=EXIT_ERROR)
        else:
            with pytest.raises(Cache.EncryptionMethodMismatch):
                self.cmd(f"--repo={self.repository_location}", "create", "test.2", "input")

    def test_repository_swap_detection2(self):
        self.create_test_files()
        self.cmd(f"--repo={self.repository_location}_unencrypted", "rcreate", "--encryption=none")
        os.environ["BORG_PASSPHRASE"] = "passphrase"
        self.cmd(f"--repo={self.repository_location}_encrypted", "rcreate", RK_ENCRYPTION)
        self.cmd(f"--repo={self.repository_location}_encrypted", "create", "test", "input")
        shutil.rmtree(self.repository_path + "_encrypted")
        os.replace(self.repository_path + "_unencrypted", self.repository_path + "_encrypted")
        if self.FORK_DEFAULT:
            self.cmd(f"--repo={self.repository_location}_encrypted", "create", "test.2", "input", exit_code=EXIT_ERROR)
        else:
            with pytest.raises(Cache.RepositoryAccessAborted):
                self.cmd(f"--repo={self.repository_location}_encrypted", "create", "test.2", "input")

    def test_repository_swap_detection_no_cache(self):
        self.create_test_files()
        os.environ["BORG_PASSPHRASE"] = "passphrase"
        self.cmd(f"--repo={self.repository_location}", "rcreate", RK_ENCRYPTION)
        repository_id = self._extract_repository_id(self.repository_path)
        self.cmd(f"--repo={self.repository_location}", "create", "test", "input")
        shutil.rmtree(self.repository_path)
        self.cmd(f"--repo={self.repository_location}", "rcreate", "--encryption=none")
        self._set_repository_id(self.repository_path, repository_id)
        self.assert_equal(repository_id, self._extract_repository_id(self.repository_path))
        self.cmd(f"--repo={self.repository_location}", "rdelete", "--cache-only")
        if self.FORK_DEFAULT:
            self.cmd(f"--repo={self.repository_location}", "create", "test.2", "input", exit_code=EXIT_ERROR)
        else:
            with pytest.raises(Cache.EncryptionMethodMismatch):
                self.cmd(f"--repo={self.repository_location}", "create", "test.2", "input")

    def test_repository_swap_detection2_no_cache(self):
        self.create_test_files()
        self.cmd(f"--repo={self.repository_location}_unencrypted", "rcreate", "--encryption=none")
        os.environ["BORG_PASSPHRASE"] = "passphrase"
        self.cmd(f"--repo={self.repository_location}_encrypted", "rcreate", RK_ENCRYPTION)
        self.cmd(f"--repo={self.repository_location}_encrypted", "create", "test", "input")
        self.cmd(f"--repo={self.repository_location}_unencrypted", "rdelete", "--cache-only")
        self.cmd(f"--repo={self.repository_location}_encrypted", "rdelete", "--cache-only")
        shutil.rmtree(self.repository_path + "_encrypted")
        os.replace(self.repository_path + "_unencrypted", self.repository_path + "_encrypted")
        if self.FORK_DEFAULT:
            self.cmd(f"--repo={self.repository_location}_encrypted", "create", "test.2", "input", exit_code=EXIT_ERROR)
        else:
            with pytest.raises(Cache.RepositoryAccessAborted):
                self.cmd(f"--repo={self.repository_location}_encrypted", "create", "test.2", "input")

    def test_repository_swap_detection_repokey_blank_passphrase(self):
        # Check that a repokey repo with a blank passphrase is considered like a plaintext repo.
        self.create_test_files()
        # User initializes her repository with her passphrase
        self.cmd(f"--repo={self.repository_location}", "rcreate", RK_ENCRYPTION)
        self.cmd(f"--repo={self.repository_location}", "create", "test", "input")
        # Attacker replaces it with her own repository, which is encrypted but has no passphrase set
        shutil.rmtree(self.repository_path)
        with environment_variable(BORG_PASSPHRASE=""):
            self.cmd(f"--repo={self.repository_location}", "rcreate", RK_ENCRYPTION)
            # Delete cache & security database, AKA switch to user perspective
            self.cmd(f"--repo={self.repository_location}", "rdelete", "--cache-only")
            shutil.rmtree(self.get_security_dir())
        with environment_variable(BORG_PASSPHRASE=None):
            # This is the part were the user would be tricked, e.g. she assumes that BORG_PASSPHRASE
            # is set, while it isn't. Previously this raised no warning,
            # since the repository is, technically, encrypted.
            if self.FORK_DEFAULT:
                self.cmd(f"--repo={self.repository_location}", "create", "test.2", "input", exit_code=EXIT_ERROR)
            else:
                with pytest.raises(Cache.CacheInitAbortedError):
                    self.cmd(f"--repo={self.repository_location}", "create", "test.2", "input")

    def test_repository_move(self):
        self.cmd(f"--repo={self.repository_location}", "rcreate", RK_ENCRYPTION)
        security_dir = self.get_security_dir()
        os.replace(self.repository_path, self.repository_path + "_new")
        with environment_variable(BORG_RELOCATED_REPO_ACCESS_IS_OK="yes"):
            self.cmd(f"--repo={self.repository_location}_new", "rinfo")
        with open(os.path.join(security_dir, "location")) as fd:
            location = fd.read()
            assert location == Location(self.repository_location + "_new").canonical_path()
        # Needs no confirmation anymore
        self.cmd(f"--repo={self.repository_location}_new", "rinfo")
        shutil.rmtree(self.cache_path)
        self.cmd(f"--repo={self.repository_location}_new", "rinfo")
        shutil.rmtree(security_dir)
        self.cmd(f"--repo={self.repository_location}_new", "rinfo")
        for file in ("location", "key-type", "manifest-timestamp"):
            assert os.path.exists(os.path.join(security_dir, file))

    def test_security_dir_compat(self):
        self.cmd(f"--repo={self.repository_location}", "rcreate", RK_ENCRYPTION)
        with open(os.path.join(self.get_security_dir(), "location"), "w") as fd:
            fd.write("something outdated")
        # This is fine, because the cache still has the correct information. security_dir and cache can disagree
        # if older versions are used to confirm a renamed repository.
        self.cmd(f"--repo={self.repository_location}", "rinfo")

    def test_unknown_unencrypted(self):
        self.cmd(f"--repo={self.repository_location}", "rcreate", "--encryption=none")
        # Ok: repository is known
        self.cmd(f"--repo={self.repository_location}", "rinfo")

        # Ok: repository is still known (through security_dir)
        shutil.rmtree(self.cache_path)
        self.cmd(f"--repo={self.repository_location}", "rinfo")

        # Needs confirmation: cache and security dir both gone (eg. another host or rm -rf ~)
        shutil.rmtree(self.cache_path)
        shutil.rmtree(self.get_security_dir())
        if self.FORK_DEFAULT:
            self.cmd(f"--repo={self.repository_location}", "rinfo", exit_code=EXIT_ERROR)
        else:
            with pytest.raises(Cache.CacheInitAbortedError):
                self.cmd(f"--repo={self.repository_location}", "rinfo")
        with environment_variable(BORG_UNKNOWN_UNENCRYPTED_REPO_ACCESS_IS_OK="yes"):
            self.cmd(f"--repo={self.repository_location}", "rinfo")

    def add_unknown_feature(self, operation):
        with Repository(self.repository_path, exclusive=True) as repository:
            manifest = Manifest.load(repository, Manifest.NO_OPERATION_CHECK)
            manifest.config["feature_flags"] = {operation.value: {"mandatory": ["unknown-feature"]}}
            manifest.write()
            repository.commit(compact=False)

    def cmd_raises_unknown_feature(self, args):
        if self.FORK_DEFAULT:
            self.cmd(*args, exit_code=EXIT_ERROR)
        else:
            with pytest.raises(MandatoryFeatureUnsupported) as excinfo:
                self.cmd(*args)
            assert excinfo.value.args == (["unknown-feature"],)

    def test_unknown_feature_on_create(self):
        print(self.cmd(f"--repo={self.repository_location}", "rcreate", RK_ENCRYPTION))
        self.add_unknown_feature(Manifest.Operation.WRITE)
        self.cmd_raises_unknown_feature([f"--repo={self.repository_location}", "create", "test", "input"])

    def test_unknown_feature_on_cache_sync(self):
        self.cmd(f"--repo={self.repository_location}", "rcreate", RK_ENCRYPTION)
        self.cmd(f"--repo={self.repository_location}", "rdelete", "--cache-only")
        self.add_unknown_feature(Manifest.Operation.READ)
        self.cmd_raises_unknown_feature([f"--repo={self.repository_location}", "create", "test", "input"])

    def test_unknown_feature_on_change_passphrase(self):
        print(self.cmd(f"--repo={self.repository_location}", "rcreate", RK_ENCRYPTION))
        self.add_unknown_feature(Manifest.Operation.CHECK)
        self.cmd_raises_unknown_feature([f"--repo={self.repository_location}", "key", "change-passphrase"])

    def test_unknown_feature_on_read(self):
        print(self.cmd(f"--repo={self.repository_location}", "rcreate", RK_ENCRYPTION))
        self.cmd(f"--repo={self.repository_location}", "create", "test", "input")
        self.add_unknown_feature(Manifest.Operation.READ)
        with changedir("output"):
            self.cmd_raises_unknown_feature([f"--repo={self.repository_location}", "extract", "test"])

        self.cmd_raises_unknown_feature([f"--repo={self.repository_location}", "rlist"])
        self.cmd_raises_unknown_feature([f"--repo={self.repository_location}", "info", "-a", "test"])

    def test_unknown_feature_on_rename(self):
        print(self.cmd(f"--repo={self.repository_location}", "rcreate", RK_ENCRYPTION))
        self.cmd(f"--repo={self.repository_location}", "create", "test", "input")
        self.add_unknown_feature(Manifest.Operation.CHECK)
        self.cmd_raises_unknown_feature([f"--repo={self.repository_location}", "rename", "test", "other"])

    def test_unknown_feature_on_delete(self):
        print(self.cmd(f"--repo={self.repository_location}", "rcreate", RK_ENCRYPTION))
        self.cmd(f"--repo={self.repository_location}", "create", "test", "input")
        self.add_unknown_feature(Manifest.Operation.DELETE)
        # delete of an archive raises
        self.cmd_raises_unknown_feature([f"--repo={self.repository_location}", "delete", "-a", "test"])
        self.cmd_raises_unknown_feature([f"--repo={self.repository_location}", "prune", "--keep-daily=3"])
        # delete of the whole repository ignores features
        self.cmd(f"--repo={self.repository_location}", "rdelete")

    @unittest.skipUnless(llfuse, "llfuse not installed")
    def test_unknown_feature_on_mount(self):
        self.cmd(f"--repo={self.repository_location}", "rcreate", RK_ENCRYPTION)
        self.cmd(f"--repo={self.repository_location}", "create", "test", "input")
        self.add_unknown_feature(Manifest.Operation.READ)
        mountpoint = os.path.join(self.tmpdir, "mountpoint")
        os.mkdir(mountpoint)
        # XXX this might hang if it doesn't raise an error
        self.cmd_raises_unknown_feature([f"--repo={self.repository_location}::test", "mount", mountpoint])

    @pytest.mark.allow_cache_wipe
    def test_unknown_mandatory_feature_in_cache(self):
        remote_repo = bool(self.prefix)
        print(self.cmd(f"--repo={self.repository_location}", "rcreate", RK_ENCRYPTION))

        with Repository(self.repository_path, exclusive=True) as repository:
            if remote_repo:
                repository._location = Location(self.repository_location)
            manifest = Manifest.load(repository, Manifest.NO_OPERATION_CHECK)
            with Cache(repository, manifest) as cache:
                cache.begin_txn()
                cache.cache_config.mandatory_features = {"unknown-feature"}
                cache.commit()

        if self.FORK_DEFAULT:
            self.cmd(f"--repo={self.repository_location}", "create", "test", "input")
        else:
            called = False
            wipe_cache_safe = LocalCache.wipe_cache

            def wipe_wrapper(*args):
                nonlocal called
                called = True
                wipe_cache_safe(*args)

            with patch.object(LocalCache, "wipe_cache", wipe_wrapper):
                self.cmd(f"--repo={self.repository_location}", "create", "test", "input")

            assert called

        with Repository(self.repository_path, exclusive=True) as repository:
            if remote_repo:
                repository._location = Location(self.repository_location)
            manifest = Manifest.load(repository, Manifest.NO_OPERATION_CHECK)
            with Cache(repository, manifest) as cache:
                assert cache.cache_config.mandatory_features == set()

    def test_check_cache(self):
        self.cmd(f"--repo={self.repository_location}", "rcreate", RK_ENCRYPTION)
        self.cmd(f"--repo={self.repository_location}", "create", "test", "input")
        with self.open_repository() as repository:
            manifest = Manifest.load(repository, Manifest.NO_OPERATION_CHECK)
            with Cache(repository, manifest, sync=False) as cache:
                cache.begin_txn()
                cache.chunks.incref(list(cache.chunks.iteritems())[0][0])
                cache.commit()
        with pytest.raises(AssertionError):
            self.check_cache()


class ManifestAuthenticationTest(ArchiverTestCaseBase):
    def spoof_manifest(self, repository):
        with repository:
            manifest = Manifest.load(repository, Manifest.NO_OPERATION_CHECK)
            cdata = manifest.repo_objs.format(
                Manifest.MANIFEST_ID,
                {},
                msgpack.packb(
                    {
                        "version": 1,
                        "archives": {},
                        "config": {},
                        "timestamp": (datetime.now(tz=timezone.utc) + timedelta(days=1)).isoformat(
                            timespec="microseconds"
                        ),
                    }
                ),
            )
            repository.put(Manifest.MANIFEST_ID, cdata)
            repository.commit(compact=False)

    def test_fresh_init_tam_required(self):
        self.cmd(f"--repo={self.repository_location}", "rcreate", RK_ENCRYPTION)
        repository = Repository(self.repository_path, exclusive=True)
        with repository:
            manifest = Manifest.load(repository, Manifest.NO_OPERATION_CHECK)
            cdata = manifest.repo_objs.format(
                Manifest.MANIFEST_ID,
                {},
                msgpack.packb(
                    {
                        "version": 1,
                        "archives": {},
                        "timestamp": (datetime.now(tz=timezone.utc) + timedelta(days=1)).isoformat(
                            timespec="microseconds"
                        ),
                    }
                ),
            )
            repository.put(Manifest.MANIFEST_ID, cdata)
            repository.commit(compact=False)

        with pytest.raises(TAMRequiredError):
            self.cmd(f"--repo={self.repository_location}", "rlist")

    def test_not_required(self):
        self.cmd(f"--repo={self.repository_location}", "rcreate", RK_ENCRYPTION)
        self.create_src_archive("archive1234")
        repository = Repository(self.repository_path, exclusive=True)
        # Manifest must be authenticated now
        output = self.cmd(f"--repo={self.repository_location}", "rlist", "--debug")
        assert "archive1234" in output
        assert "TAM-verified manifest" in output
        # Try to spoof / modify pre-1.0.9
        self.spoof_manifest(repository)
        # Fails
        with pytest.raises(TAMRequiredError):
            self.cmd(f"--repo={self.repository_location}", "rlist")


class RemoteArchiverTestCase(RemoteArchiverTestCaseBase, ArchiverTestCase):
    def test_remote_repo_restrict_to_path(self):
        # restricted to repo directory itself:
        with patch.object(RemoteRepository, "extra_test_args", ["--restrict-to-path", self.repository_path]):
            self.cmd(f"--repo={self.repository_location}", "rcreate", RK_ENCRYPTION)
        # restricted to repo directory itself, fail for other directories with same prefix:
        with patch.object(RemoteRepository, "extra_test_args", ["--restrict-to-path", self.repository_path]):
            with pytest.raises(PathNotAllowed):
                self.cmd(f"--repo={self.repository_location}_0", "rcreate", RK_ENCRYPTION)

        # restricted to a completely different path:
        with patch.object(RemoteRepository, "extra_test_args", ["--restrict-to-path", "/foo"]):
            with pytest.raises(PathNotAllowed):
                self.cmd(f"--repo={self.repository_location}_1", "rcreate", RK_ENCRYPTION)
        path_prefix = os.path.dirname(self.repository_path)
        # restrict to repo directory's parent directory:
        with patch.object(RemoteRepository, "extra_test_args", ["--restrict-to-path", path_prefix]):
            self.cmd(f"--repo={self.repository_location}_2", "rcreate", RK_ENCRYPTION)
        # restrict to repo directory's parent directory and another directory:
        with patch.object(
            RemoteRepository, "extra_test_args", ["--restrict-to-path", "/foo", "--restrict-to-path", path_prefix]
        ):
            self.cmd(f"--repo={self.repository_location}_3", "rcreate", RK_ENCRYPTION)

    def test_remote_repo_restrict_to_repository(self):
        # restricted to repo directory itself:
        with patch.object(RemoteRepository, "extra_test_args", ["--restrict-to-repository", self.repository_path]):
            self.cmd(f"--repo={self.repository_location}", "rcreate", RK_ENCRYPTION)
        parent_path = os.path.join(self.repository_path, "..")
        with patch.object(RemoteRepository, "extra_test_args", ["--restrict-to-repository", parent_path]):
            with pytest.raises(PathNotAllowed):
                self.cmd(f"--repo={self.repository_location}", "rcreate", RK_ENCRYPTION)

    def test_remote_repo_strip_components_doesnt_leak(self):
        self.cmd(f"--repo={self.repository_location}", "rcreate", RK_ENCRYPTION)
        self.create_regular_file("dir/file", contents=b"test file contents 1")
        self.create_regular_file("dir/file2", contents=b"test file contents 2")
        self.create_regular_file("skipped-file1", contents=b"test file contents 3")
        self.create_regular_file("skipped-file2", contents=b"test file contents 4")
        self.create_regular_file("skipped-file3", contents=b"test file contents 5")
        self.cmd(f"--repo={self.repository_location}", "create", "test", "input")
        marker = "cached responses left in RemoteRepository"
        with changedir("output"):
            res = self.cmd(
                f"--repo={self.repository_location}", "extract", "test", "--debug", "--strip-components", "3"
            )
            assert marker not in res
            with self.assert_creates_file("file"):
                res = self.cmd(
                    f"--repo={self.repository_location}", "extract", "test", "--debug", "--strip-components", "2"
                )
                assert marker not in res
            with self.assert_creates_file("dir/file"):
                res = self.cmd(
                    f"--repo={self.repository_location}", "extract", "test", "--debug", "--strip-components", "1"
                )
                assert marker not in res
            with self.assert_creates_file("input/dir/file"):
                res = self.cmd(
                    f"--repo={self.repository_location}", "extract", "test", "--debug", "--strip-components", "0"
                )
                assert marker not in res



import os
import unittest

from ...constants import *  # NOQA
from . import ArchiverTestCaseBase, RemoteArchiverTestCaseBase, ArchiverTestCaseBinaryBase, RK_ENCRYPTION, BORG_EXES


class ArchiverTestCase(ArchiverTestCaseBase):
    def test_delete_repo(self):
        self.create_regular_file("file1", size=1024 * 80)
        self.create_regular_file("dir2/file2", size=1024 * 80)
        self.cmd(f"--repo={self.repository_location}", "rcreate", RK_ENCRYPTION)
        self.cmd(f"--repo={self.repository_location}", "create", "test", "input")
        self.cmd(f"--repo={self.repository_location}", "create", "test.2", "input")
        os.environ["BORG_DELETE_I_KNOW_WHAT_I_AM_DOING"] = "no"
        self.cmd(f"--repo={self.repository_location}", "rdelete", exit_code=2)
        assert os.path.exists(self.repository_path)
        os.environ["BORG_DELETE_I_KNOW_WHAT_I_AM_DOING"] = "YES"
        self.cmd(f"--repo={self.repository_location}", "rdelete")
        # Make sure the repo is gone
        self.assertFalse(os.path.exists(self.repository_path))


class RemoteArchiverTestCase(RemoteArchiverTestCaseBase, ArchiverTestCase):
    """run the same tests, but with a remote repository"""


@unittest.skipUnless("binary" in BORG_EXES, "no borg.exe available")
class ArchiverTestCaseBinary(ArchiverTestCaseBinaryBase, ArchiverTestCase):
    """runs the same tests, but via the borg binary"""



import argparse
import fnmatch
import os.path
import re
import sys
import unicodedata
from collections import namedtuple
from enum import Enum

from .helpers import clean_lines, shellpattern
from .helpers.errors import Error


def parse_patternfile_line(line, roots, ie_commands, fallback):
    """Parse a pattern-file line and act depending on which command it represents."""
    ie_command = parse_inclexcl_command(line, fallback=fallback)
    if ie_command.cmd is IECommand.RootPath:
        roots.append(ie_command.val)
    elif ie_command.cmd is IECommand.PatternStyle:
        fallback = ie_command.val
    else:
        # it is some kind of include/exclude command
        ie_commands.append(ie_command)
    return fallback


def load_pattern_file(fileobj, roots, ie_commands, fallback=None):
    if fallback is None:
        fallback = ShellPattern  # ShellPattern is defined later in this module
    for line in clean_lines(fileobj):
        fallback = parse_patternfile_line(line, roots, ie_commands, fallback)


def load_exclude_file(fileobj, patterns):
    for patternstr in clean_lines(fileobj):
        patterns.append(parse_exclude_pattern(patternstr))


class ArgparsePatternAction(argparse.Action):
    def __init__(self, nargs=1, **kw):
        super().__init__(nargs=nargs, **kw)

    def __call__(self, parser, args, values, option_string=None):
        parse_patternfile_line(values[0], args.paths, args.patterns, ShellPattern)


class ArgparsePatternFileAction(argparse.Action):
    def __init__(self, nargs=1, **kw):
        super().__init__(nargs=nargs, **kw)

    def __call__(self, parser, args, values, option_string=None):
        """Load and parse patterns from a file.
        Lines empty or starting with '#' after stripping whitespace on both line ends are ignored.
        """
        filename = values[0]
        try:
            with open(filename) as f:
                self.parse(f, args)
        except FileNotFoundError as e:
            raise Error(str(e))

    def parse(self, fobj, args):
        load_pattern_file(fobj, args.paths, args.patterns)


class ArgparseExcludeFileAction(ArgparsePatternFileAction):
    def parse(self, fobj, args):
        load_exclude_file(fobj, args.patterns)


class PatternMatcher:
    """Represents a collection of pattern objects to match paths against.

    *fallback* is a boolean value that *match()* returns if no matching patterns are found.

    """

    def __init__(self, fallback=None):
        self._items = []

        # Value to return from match function when none of the patterns match.
        self.fallback = fallback

        # optimizations
        self._path_full_patterns = {}  # full path -> return value

        # indicates whether the last match() call ended on a pattern for which
        # we should recurse into any matching folder.  Will be set to True or
        # False when calling match().
        self.recurse_dir = None

        # whether to recurse into directories when no match is found
        # TODO: allow modification as a config option?
        self.recurse_dir_default = True

        self.include_patterns = []

        # TODO: move this info to parse_inclexcl_command and store in PatternBase subclass?
        self.is_include_cmd = {IECommand.Exclude: False, IECommand.ExcludeNoRecurse: False, IECommand.Include: True}

    def empty(self):
        return not len(self._items) and not len(self._path_full_patterns)

    def _add(self, pattern, cmd):
        """*cmd* is an IECommand value."""
        if isinstance(pattern, PathFullPattern):
            key = pattern.pattern  # full, normalized path
            self._path_full_patterns[key] = cmd
        else:
            self._items.append((pattern, cmd))

    def add(self, patterns, cmd):
        """Add list of patterns to internal list. *cmd* indicates whether the
        pattern is an include/exclude pattern, and whether recursion should be
        done on excluded folders.
        """
        for pattern in patterns:
            self._add(pattern, cmd)

    def add_includepaths(self, include_paths):
        """Used to add inclusion-paths from args.paths (from commandline)."""
        include_patterns = [parse_pattern(p, PathPrefixPattern) for p in include_paths]
        self.add(include_patterns, IECommand.Include)
        self.fallback = not include_patterns
        self.include_patterns = include_patterns

    def get_unmatched_include_patterns(self):
        "Note that this only returns patterns added via *add_includepaths*."
        return [p for p in self.include_patterns if p.match_count == 0]

    def add_inclexcl(self, patterns):
        """Add list of patterns (of type CmdTuple) to internal list."""
        for pattern, cmd in patterns:
            self._add(pattern, cmd)

    def match(self, path):
        """Return True or False depending on whether *path* is matched.

        If no match is found among the patterns in this matcher, then the value
        in self.fallback is returned (defaults to None).

        """
        path = normalize_path(path).lstrip(os.path.sep)
        # do a fast lookup for full path matches (note: we do not count such matches):
        non_existent = object()
        value = self._path_full_patterns.get(path, non_existent)

        if value is not non_existent:
            # we have a full path match!
            self.recurse_dir = command_recurses_dir(value)
            return self.is_include_cmd[value]

        # this is the slow way, if we have many patterns in self._items:
        for pattern, cmd in self._items:
            if pattern.match(path, normalize=False):
                self.recurse_dir = pattern.recurse_dir
                return self.is_include_cmd[cmd]

        # by default we will recurse if there is no match
        self.recurse_dir = self.recurse_dir_default
        return self.fallback


def normalize_path(path):
    """normalize paths for MacOS (but do nothing on other platforms)"""
    # HFS+ converts paths to a canonical form, so users shouldn't be required to enter an exact match.
    # Windows and Unix filesystems allow different forms, so users always have to enter an exact match.
    return unicodedata.normalize("NFD", path) if sys.platform == "darwin" else path


class PatternBase:
    """Shared logic for inclusion/exclusion patterns."""

    PREFIX: str = None

    def __init__(self, pattern, recurse_dir=False):
        self.pattern_orig = pattern
        self.match_count = 0
        pattern = normalize_path(pattern)
        self._prepare(pattern)
        self.recurse_dir = recurse_dir

    def match(self, path, normalize=True):
        """Return a boolean indicating whether *path* is matched by this pattern.

        If normalize is True (default), the path will get normalized using normalize_path(),
        otherwise it is assumed that it already is normalized using that function.
        """
        if normalize:
            path = normalize_path(path)
        matches = self._match(path)
        if matches:
            self.match_count += 1
        return matches

    def __repr__(self):
        return f"{type(self)}({self.pattern})"

    def __str__(self):
        return self.pattern_orig

    def _prepare(self, pattern):
        "Should set the value of self.pattern"
        raise NotImplementedError

    def _match(self, path):
        raise NotImplementedError


class PathFullPattern(PatternBase):
    """Full match of a path."""

    PREFIX = "pf"

    def _prepare(self, pattern):
        self.pattern = os.path.normpath(pattern).lstrip(os.path.sep)  # sep at beginning is removed

    def _match(self, path):
        return path == self.pattern


# For PathPrefixPattern, FnmatchPattern and ShellPattern, we require that the pattern either match the whole path
# or an initial segment of the path up to but not including a path separator. To unify the two cases, we add a path
# separator to the end of the path before matching.


class PathPrefixPattern(PatternBase):
    """Literal files or directories listed on the command line
    for some operations (e.g. extract, but not create).
    If a directory is specified, all paths that start with that
    path match as well.  A trailing slash makes no difference.
    """

    PREFIX = "pp"

    def _prepare(self, pattern):
        sep = os.path.sep

        self.pattern = (os.path.normpath(pattern).rstrip(sep) + sep).lstrip(sep)  # sep at beginning is removed

    def _match(self, path):
        return (path + os.path.sep).startswith(self.pattern)


class FnmatchPattern(PatternBase):
    """Shell glob patterns to exclude.  A trailing slash means to
    exclude the contents of a directory, but not the directory itself.
    """

    PREFIX = "fm"

    def _prepare(self, pattern):
        if pattern.endswith(os.path.sep):
            pattern = os.path.normpath(pattern).rstrip(os.path.sep) + os.path.sep + "*" + os.path.sep
        else:
            pattern = os.path.normpath(pattern) + os.path.sep + "*"

        self.pattern = pattern.lstrip(os.path.sep)  # sep at beginning is removed

        # fnmatch and re.match both cache compiled regular expressions.
        # Nevertheless, this is about 10 times faster.
        self.regex = re.compile(fnmatch.translate(self.pattern))

    def _match(self, path):
        return self.regex.match(path + os.path.sep) is not None


class ShellPattern(PatternBase):
    """Shell glob patterns to exclude.  A trailing slash means to
    exclude the contents of a directory, but not the directory itself.
    """

    PREFIX = "sh"

    def _prepare(self, pattern):
        sep = os.path.sep

        if pattern.endswith(sep):
            pattern = os.path.normpath(pattern).rstrip(sep) + sep + "**" + sep + "*" + sep
        else:
            pattern = os.path.normpath(pattern) + sep + "**" + sep + "*"

        self.pattern = pattern.lstrip(sep)  # sep at beginning is removed
        self.regex = re.compile(shellpattern.translate(self.pattern))

    def _match(self, path):
        return self.regex.match(path + os.path.sep) is not None


class RegexPattern(PatternBase):
    """Regular expression to exclude."""

    PREFIX = "re"

    def _prepare(self, pattern):
        self.pattern = pattern  # sep at beginning is NOT removed
        self.regex = re.compile(pattern)

    def _match(self, path):
        # Normalize path separators
        if os.path.sep != "/":
            path = path.replace(os.path.sep, "/")

        return self.regex.search(path) is not None


_PATTERN_CLASSES = {FnmatchPattern, PathFullPattern, PathPrefixPattern, RegexPattern, ShellPattern}

_PATTERN_CLASS_BY_PREFIX = {i.PREFIX: i for i in _PATTERN_CLASSES}

CmdTuple = namedtuple("CmdTuple", "val cmd")


class IECommand(Enum):
    """A command that an InclExcl file line can represent."""

    RootPath = 1
    PatternStyle = 2
    Include = 3
    Exclude = 4
    ExcludeNoRecurse = 5


def command_recurses_dir(cmd):
    # TODO?: raise error or return None if *cmd* is RootPath or PatternStyle
    return cmd not in [IECommand.ExcludeNoRecurse]


def get_pattern_class(prefix):
    try:
        return _PATTERN_CLASS_BY_PREFIX[prefix]
    except KeyError:
        raise ValueError(f"Unknown pattern style: {prefix}") from None


def parse_pattern(pattern, fallback=FnmatchPattern, recurse_dir=True):
    """Read pattern from string and return an instance of the appropriate implementation class."""
    if len(pattern) > 2 and pattern[2] == ":" and pattern[:2].isalnum():
        (style, pattern) = (pattern[:2], pattern[3:])
        cls = get_pattern_class(style)
    else:
        cls = fallback
    return cls(pattern, recurse_dir)


def parse_exclude_pattern(pattern_str, fallback=FnmatchPattern):
    """Read pattern from string and return an instance of the appropriate implementation class."""
    epattern_obj = parse_pattern(pattern_str, fallback, recurse_dir=False)
    return CmdTuple(epattern_obj, IECommand.ExcludeNoRecurse)


def parse_inclexcl_command(cmd_line_str, fallback=ShellPattern):
    """Read a --patterns-from command from string and return a CmdTuple object."""

    cmd_prefix_map = {
        "-": IECommand.Exclude,
        "!": IECommand.ExcludeNoRecurse,
        "+": IECommand.Include,
        "R": IECommand.RootPath,
        "r": IECommand.RootPath,
        "P": IECommand.PatternStyle,
        "p": IECommand.PatternStyle,
    }
    if not cmd_line_str:
        raise argparse.ArgumentTypeError("A pattern/command must not be empty.")

    cmd = cmd_prefix_map.get(cmd_line_str[0])
    if cmd is None:
        raise argparse.ArgumentTypeError("A pattern/command must start with anyone of: %s" % ", ".join(cmd_prefix_map))

    # remaining text on command-line following the command character
    remainder_str = cmd_line_str[1:].lstrip()
    if not remainder_str:
        raise argparse.ArgumentTypeError("A pattern/command must have a value part.")

    if cmd is IECommand.RootPath:
        # TODO: validate string?
        val = remainder_str
    elif cmd is IECommand.PatternStyle:
        # then remainder_str is something like 're' or 'sh'
        try:
            val = get_pattern_class(remainder_str)
        except ValueError:
            raise argparse.ArgumentTypeError(f"Invalid pattern style: {remainder_str}")
    else:
        # determine recurse_dir based on command type
        recurse_dir = command_recurses_dir(cmd)
        val = parse_pattern(remainder_str, fallback, recurse_dir)

    return CmdTuple(val, cmd)


def get_regex_from_pattern(pattern: str) -> str:
    """
    return a regular expression string corresponding to the given pattern string.

    the allowed pattern types are similar to the ones implemented by PatternBase subclasses,
    but here we rather do generic string matching, not specialised filesystem paths matching.
    """
    if len(pattern) > 2 and pattern[2] == ":" and pattern[:2] in {"sh", "re", "id"}:
        (style, pattern) = (pattern[:2], pattern[3:])
    else:
        (style, pattern) = ("id", pattern)  # "identical" match is the default
    if style == "sh":
        # (?ms) (meaning re.MULTILINE and re.DOTALL) are not desired here.
        regex = shellpattern.translate(pattern, match_end="").removeprefix("(?ms)")
    elif style == "re":
        regex = pattern
    elif style == "id":
        regex = re.escape(pattern)
    else:
        raise NotImplementedError
    return regex



import argparse
from collections import defaultdict

from ._common import with_repository
from ..constants import *  # NOQA
from ..compress import CompressionSpec, ObfuscateSize, Auto, COMPRESSOR_TABLE
from ..helpers import sig_int, ProgressIndicatorPercent

from ..manifest import Manifest

from ..logger import create_logger

logger = create_logger()


def find_chunks(repository, repo_objs, stats, ctype, clevel, olevel):
    """find chunks that need processing (usually: recompression)."""
    # to do it this way is maybe not obvious, thus keeping the essential design criteria here:
    # - determine the chunk ids at one point in time (== do a **full** scan in one go) **before**
    # writing to the repo (and especially before doing a compaction, which moves segment files around)
    # - get the chunk ids in **on-disk order** (so we can efficiently compact while processing the chunks)
    # - only put the ids into the list that actually need recompression (keeps it a little shorter in some cases)
    recompress_ids = []
    compr_keys = stats["compr_keys"] = set()
    compr_wanted = ctype, clevel, olevel
    state = None
    chunks_count = len(repository)
    chunks_limit = min(1000, max(100, chunks_count // 1000))
    pi = ProgressIndicatorPercent(
        total=chunks_count,
        msg="Searching for recompression candidates %3.1f%%",
        step=0.1,
        msgid="rcompress.find_chunks",
    )
    while True:
        chunk_ids, state = repository.scan(limit=chunks_limit, state=state)
        if not chunk_ids:
            break
        for id, chunk_no_data in zip(chunk_ids, repository.get_many(chunk_ids, read_data=False)):
            meta = repo_objs.parse_meta(id, chunk_no_data)
            compr_found = meta["ctype"], meta["clevel"], meta.get("olevel", -1)
            if compr_found != compr_wanted:
                recompress_ids.append(id)
            compr_keys.add(compr_found)
            stats[compr_found] += 1
            stats["checked_count"] += 1
            pi.show(increase=1)
    pi.finish()
    return recompress_ids


def process_chunks(repository, repo_objs, stats, recompress_ids, olevel):
    """process some chunks (usually: recompress)"""
    compr_keys = stats["compr_keys"]
    if compr_keys == 0:  # work around defaultdict(int)
        compr_keys = stats["compr_keys"] = set()
    for id, chunk in zip(recompress_ids, repository.get_many(recompress_ids, read_data=True)):
        old_size = len(chunk)
        stats["old_size"] += old_size
        meta, data = repo_objs.parse(id, chunk)
        compr_old = meta["ctype"], meta["clevel"], meta.get("olevel", -1)
        if olevel == -1:
            # if the chunk was obfuscated, but should not be in future, remove related metadata
            meta.pop("olevel", None)
            meta.pop("psize", None)
        chunk = repo_objs.format(id, meta, data)
        compr_done = meta["ctype"], meta["clevel"], meta.get("olevel", -1)
        if compr_done != compr_old:
            # we actually changed something
            repository.put(id, chunk, wait=False)
            repository.async_response(wait=False)
            stats["new_size"] += len(chunk)
            compr_keys.add(compr_done)
            stats[compr_done] += 1
            stats["recompressed_count"] += 1
        else:
            # It might be that the old chunk used compression none or lz4 (for whatever reason,
            # including the old compressor being a DecidingCompressor) AND we used a
            # DecidingCompressor now, which did NOT compress like we wanted, but decided
            # to use the same compression (and obfuscation) we already had.
            # In this case, we just keep the old chunk and do not rewrite it -
            # This is important to avoid rewriting such chunks **again and again**.
            stats["new_size"] += old_size
            compr_keys.add(compr_old)
            stats[compr_old] += 1
            stats["kept_count"] += 1


def format_compression_spec(ctype, clevel, olevel):
    obfuscation = "" if olevel == -1 else f"obfuscate,{olevel},"
    for cname, cls in COMPRESSOR_TABLE.items():
        if cls.ID == ctype:
            cname = f"{cname}"
            break
    else:
        cname = f"{ctype}"
    clevel = f",{clevel}" if clevel != 255 else ""
    return obfuscation + cname + clevel


class RCompressMixIn:
    @with_repository(cache=False, manifest=True, exclusive=True, compatibility=(Manifest.Operation.CHECK,))
    def do_rcompress(self, args, repository, manifest):
        """Repository (re-)compression"""

        def get_csettings(c):
            if isinstance(c, Auto):
                return get_csettings(c.compressor)
            if isinstance(c, ObfuscateSize):
                ctype, clevel, _ = get_csettings(c.compressor)
                olevel = c.level
                return ctype, clevel, olevel
            ctype, clevel, olevel = c.ID, c.level, -1
            return ctype, clevel, olevel

        repo_objs = manifest.repo_objs
        ctype, clevel, olevel = get_csettings(repo_objs.compressor)  # desired compression set by --compression

        def checkpoint_func():
            while repository.async_response(wait=True) is not None:
                pass
            repository.commit(compact=True)

        stats_find = defaultdict(int)
        stats_process = defaultdict(int)
        recompress_ids = find_chunks(repository, repo_objs, stats_find, ctype, clevel, olevel)
        recompress_candidate_count = len(recompress_ids)
        chunks_limit = min(1000, max(100, recompress_candidate_count // 1000))
        uncommitted_chunks = 0

        # start a new transaction
        data = repository.get(Manifest.MANIFEST_ID)
        repository.put(Manifest.MANIFEST_ID, data)
        uncommitted_chunks += 1

        pi = ProgressIndicatorPercent(
            total=len(recompress_ids), msg="Recompressing %3.1f%%", step=0.1, msgid="rcompress.process_chunks"
        )
        while recompress_ids:
            if sig_int and sig_int.action_done():
                break
            ids, recompress_ids = recompress_ids[:chunks_limit], recompress_ids[chunks_limit:]
            process_chunks(repository, repo_objs, stats_process, ids, olevel)
            pi.show(increase=len(ids))
            checkpointed = self.maybe_checkpoint(
                checkpoint_func=checkpoint_func, checkpoint_interval=args.checkpoint_interval
            )
            uncommitted_chunks = 0 if checkpointed else (uncommitted_chunks + len(ids))
        pi.finish()
        if sig_int:
            # Ctrl-C / SIGINT: do not checkpoint (commit) again, we already have a checkpoint in this case.
            self.print_error("Got Ctrl-C / SIGINT.")
        elif uncommitted_chunks > 0:
            checkpoint_func()
        if args.stats:
            print()
            print("Recompression stats:")
            print(f"Size: previously {stats_process['old_size']} -> now {stats_process['new_size']} bytes.")
            print(
                f"Change: "
                f"{stats_process['new_size'] - stats_process['old_size']} bytes == "
                f"{100.0 * stats_process['new_size'] / stats_process['old_size']:3.2f}%"
            )
            print("Found chunks stats (before processing):")
            for ck in stats_find["compr_keys"]:
                pretty_ck = format_compression_spec(*ck)
                print(f"{pretty_ck}: {stats_find[ck]}")
            print(f"Total: {stats_find['checked_count']}")

            print(f"Candidates for recompression: {recompress_candidate_count}")

            print("Processed chunks stats (after processing):")
            for ck in stats_process["compr_keys"]:
                pretty_ck = format_compression_spec(*ck)
                print(f"{pretty_ck}: {stats_process[ck]}")
            print(f"Recompressed and rewritten: {stats_process['recompressed_count']}")
            print(f"Kept as is: {stats_process['kept_count']}")
            print(f"Total: {stats_process['recompressed_count'] + stats_process['kept_count']}")

        return self.exit_code

    def build_parser_rcompress(self, subparsers, common_parser, mid_common_parser):
        from ._common import process_epilog

        rcompress_epilog = process_epilog(
            """
        Repository (re-)compression (and/or re-obfuscation).

        Reads all chunks in the repository (in on-disk order, this is important for
        compaction) and recompresses them if they are not already using the compression
        type/level and obfuscation level given via ``--compression``.

        If the outcome of the chunk processing indicates a change in compression
        type/level or obfuscation level, the processed chunk is written to the repository.
        Please note that the outcome might not always be the desired compression
        type/level - if no compression gives a shorter output, that might be chosen.

        Every ``--checkpoint-interval``, progress is committed to the repository and
        the repository is compacted (this is to keep temporary repo space usage in bounds).
        A lower checkpoint interval means lower temporary repo space usage, but also
        slower progress due to higher overhead (and vice versa).

        Please note that this command can not work in low (or zero) free disk space
        conditions.

        If the ``borg rcompress`` process receives a SIGINT signal (Ctrl-C), the repo
        will be committed and compacted and borg will terminate cleanly afterwards.

        Both ``--progress`` and ``--stats`` are recommended when ``borg rcompress``
        is used interactively.

        You do **not** need to run ``borg compact`` after ``borg rcompress``.
        """
        )
        subparser = subparsers.add_parser(
            "rcompress",
            parents=[common_parser],
            add_help=False,
            description=self.do_rcompress.__doc__,
            epilog=rcompress_epilog,
            formatter_class=argparse.RawDescriptionHelpFormatter,
            help=self.do_rcompress.__doc__,
        )
        subparser.set_defaults(func=self.do_rcompress)

        subparser.add_argument(
            "-C",
            "--compression",
            metavar="COMPRESSION",
            dest="compression",
            type=CompressionSpec,
            default=CompressionSpec("lz4"),
            help="select compression algorithm, see the output of the " '"borg help compression" command for details.',
        )

        subparser.add_argument("-s", "--stats", dest="stats", action="store_true", help="print statistics")

        subparser.add_argument(
            "-c",
            "--checkpoint-interval",
            metavar="SECONDS",
            dest="checkpoint_interval",
            type=int,
            default=1800,
            help="write checkpoint every SECONDS seconds (Default: 1800)",
        )



from packaging.version import parse as parse_version

from ._version import version as __version__


_v = parse_version(__version__)
__version_tuple__ = _v._version.release  # type: ignore

# assert that all semver components are integers
# this is mainly to show errors when people repackage poorly
# and setuptools_scm determines a 0.1.dev... version
assert all(isinstance(v, int) for v in __version_tuple__), (
    """\
broken borgbackup version metadata: %r

version metadata is obtained dynamically on installation via setuptools_scm,
please ensure your git repo has the correct tags or you provide the version
using SETUPTOOLS_SCM_PRETEND_VERSION in your build script.
"""
    % __version__
)



import enum
import os
import os.path
import re
from collections import abc, namedtuple
from datetime import datetime, timedelta, timezone
from operator import attrgetter
from typing import Sequence, FrozenSet

from .logger import create_logger

logger = create_logger()

from .constants import *  # NOQA
from .helpers.datastruct import StableDict
from .helpers.parseformat import bin_to_hex
from .helpers.time import parse_timestamp, calculate_relative_offset, archive_ts_now
from .helpers.errors import Error
from .patterns import get_regex_from_pattern
from .repoobj import RepoObj


class NoManifestError(Error):
    """Repository has no manifest."""


class MandatoryFeatureUnsupported(Error):
    """Unsupported repository feature(s) {}. A newer version of borg is required to access this repository."""


ArchiveInfo = namedtuple("ArchiveInfo", "name id ts")

AI_HUMAN_SORT_KEYS = ["timestamp"] + list(ArchiveInfo._fields)
AI_HUMAN_SORT_KEYS.remove("ts")


def filter_archives_by_date(archives, older=None, newer=None, oldest=None, newest=None):
    def get_first_and_last_archive_ts(archives_list):
        timestamps = [x.ts for x in archives_list]
        return min(timestamps), max(timestamps)

    now = archive_ts_now()
    earliest_ts, latest_ts = get_first_and_last_archive_ts(archives)

    until_ts = calculate_relative_offset(older, now, earlier=True) if older is not None else latest_ts
    from_ts = calculate_relative_offset(newer, now, earlier=True) if newer is not None else earliest_ts
    archives = [x for x in archives if from_ts <= x.ts <= until_ts]

    earliest_ts, latest_ts = get_first_and_last_archive_ts(archives)
    if oldest:
        until_ts = calculate_relative_offset(oldest, earliest_ts, earlier=False)
        archives = [x for x in archives if x.ts <= until_ts]
    if newest:
        from_ts = calculate_relative_offset(newest, latest_ts, earlier=True)
        archives = [x for x in archives if x.ts >= from_ts]

    return archives


class Archives(abc.MutableMapping):
    """
    Nice wrapper around the archives dict, making sure only valid types/values get in
    and we can deal with str keys (and it internally encodes to byte keys) and either
    str timestamps or datetime timestamps.
    """

    def __init__(self):
        # key: str archive name, value: dict('id': bytes_id, 'time': str_iso_ts)
        self._archives = {}

    def __len__(self):
        return len(self._archives)

    def __iter__(self):
        return iter(self._archives)

    def __getitem__(self, name):
        assert isinstance(name, str)
        values = self._archives.get(name)
        if values is None:
            raise KeyError
        ts = parse_timestamp(values["time"])
        return ArchiveInfo(name=name, id=values["id"], ts=ts)

    def __setitem__(self, name, info):
        assert isinstance(name, str)
        assert isinstance(info, tuple)
        id, ts = info
        assert isinstance(id, bytes)
        if isinstance(ts, datetime):
            ts = ts.isoformat(timespec="microseconds")
        assert isinstance(ts, str)
        self._archives[name] = {"id": id, "time": ts}

    def __delitem__(self, name):
        assert isinstance(name, str)
        del self._archives[name]

    def list(
        self,
        *,
        consider_checkpoints=True,
        match=None,
        match_end=r"\Z",
        sort_by=(),
        reverse=False,
        first=None,
        last=None,
        older=None,
        newer=None,
        oldest=None,
        newest=None
    ):
        """
        Return list of ArchiveInfo instances according to the parameters.

        First match *match* (considering *match_end*), then filter by timestamp considering *older* and *newer*.
        Second, follow with a filter considering *oldest* and *newest*, then sort by the given *sort_by* argument.

        Apply *first* and *last* filters, and then possibly *reverse* the list.

        *sort_by* is a list of sort keys applied in reverse order.
        *newer* and *older* are relative time markers that indicate offset from now.
        *newest* and *oldest* are relative time markers that indicate offset from newest/oldest archive's timestamp.


        Note: for better robustness, all filtering / limiting parameters must default to
              "not limit / not filter", so a FULL archive list is produced by a simple .list().
              some callers EXPECT to iterate over all archives in a repo for correct operation.
        """
        if isinstance(sort_by, (str, bytes)):
            raise TypeError("sort_by must be a sequence of str")

        archives = self.values()
        regex = get_regex_from_pattern(match or "re:.*")
        regex = re.compile(regex + match_end)
        archives = [x for x in archives if regex.match(x.name) is not None]

        if any([oldest, newest, older, newer]) and len(archives) > 0:
            archives = filter_archives_by_date(archives, oldest=oldest, newest=newest, newer=newer, older=older)
        if not consider_checkpoints:
            archives = [x for x in archives if ".checkpoint" not in x.name]
        for sortkey in reversed(sort_by):
            archives.sort(key=attrgetter(sortkey))
        if first:
            archives = archives[:first]
        elif last:
            archives = archives[max(len(archives) - last, 0) :]
        if reverse:
            archives.reverse()
        return archives

    def list_considering(self, args):
        """
        get a list of archives, considering --first/last/prefix/match-archives/sort/consider-checkpoints cmdline args
        """
        name = getattr(args, "name", None)
        consider_checkpoints = getattr(args, "consider_checkpoints", None)
        if name is not None:
            raise Error(
                "Giving a specific name is incompatible with options --first, --last, -a / --match-archives, and --consider-checkpoints."
            )
        return self.list(
            sort_by=args.sort_by.split(","),
            consider_checkpoints=consider_checkpoints,
            match=args.match_archives,
            first=getattr(args, "first", None),
            last=getattr(args, "last", None),
            older=getattr(args, "older", None),
            newer=getattr(args, "newer", None),
            oldest=getattr(args, "oldest", None),
            newest=getattr(args, "newest", None),
        )

    def set_raw_dict(self, d):
        """set the dict we get from the msgpack unpacker"""
        for k, v in d.items():
            assert isinstance(k, str)
            assert isinstance(v, dict) and "id" in v and "time" in v
            self._archives[k] = v

    def get_raw_dict(self):
        """get the dict we can give to the msgpack packer"""
        return self._archives


class Manifest:
    @enum.unique
    class Operation(enum.Enum):
        # The comments here only roughly describe the scope of each feature. In the end, additions need to be
        # based on potential problems older clients could produce when accessing newer repositories and the
        # trade-offs of locking version out or still allowing access. As all older versions and their exact
        # behaviours are known when introducing new features sometimes this might not match the general descriptions
        # below.

        # The READ operation describes which features are needed to list and extract the archives safely in the
        # repository.
        READ = "read"
        # The CHECK operation is for all operations that need either to understand every detail
        # of the repository (for consistency checks and repairs) or are seldom used functions that just
        # should use the most restrictive feature set because more fine grained compatibility tracking is
        # not needed.
        CHECK = "check"
        # The WRITE operation is for adding archives. Features here ensure that older clients don't add archives
        # in an old format, or is used to lock out clients that for other reasons can no longer safely add new
        # archives.
        WRITE = "write"
        # The DELETE operation is for all operations (like archive deletion) that need a 100% correct reference
        # count and the need to be able to find all (directly and indirectly) referenced chunks of a given archive.
        DELETE = "delete"

    NO_OPERATION_CHECK: Sequence[Operation] = tuple()

    SUPPORTED_REPO_FEATURES: FrozenSet[str] = frozenset([])

    MANIFEST_ID = b"\0" * 32

    def __init__(self, key, repository, item_keys=None, ro_cls=RepoObj):
        self.archives = Archives()
        self.config = {}
        self.key = key
        self.repo_objs = ro_cls(key)
        self.repository = repository
        self.item_keys = frozenset(item_keys) if item_keys is not None else ITEM_KEYS
        self.tam_verified = False
        self.timestamp = None

    @property
    def id_str(self):
        return bin_to_hex(self.id)

    @property
    def last_timestamp(self):
        return parse_timestamp(self.timestamp)

    @classmethod
    def load(cls, repository, operations, key=None, force_tam_not_required=False, *, ro_cls=RepoObj):
        from .item import ManifestItem
        from .crypto.key import key_factory, tam_required_file, tam_required
        from .repository import Repository

        try:
            cdata = repository.get(cls.MANIFEST_ID)
        except Repository.ObjectNotFound:
            raise NoManifestError
        if not key:
            key = key_factory(repository, cdata, ro_cls=ro_cls)
        manifest = cls(key, repository, ro_cls=ro_cls)
        _, data = manifest.repo_objs.parse(cls.MANIFEST_ID, cdata)
        manifest_dict, manifest.tam_verified = key.unpack_and_verify_manifest(
            data, force_tam_not_required=force_tam_not_required
        )
        m = ManifestItem(internal_dict=manifest_dict)
        manifest.id = manifest.repo_objs.id_hash(data)
        if m.get("version") not in (1, 2):
            raise ValueError("Invalid manifest version")
        manifest.archives.set_raw_dict(m.archives)
        manifest.timestamp = m.get("timestamp")
        manifest.config = m.config
        # valid item keys are whatever is known in the repo or every key we know
        manifest.item_keys = ITEM_KEYS | frozenset(m.get("item_keys", []))

        if manifest.tam_verified:
            manifest_required = manifest.config.get("tam_required", False)
            security_required = tam_required(repository)
            if manifest_required and not security_required:
                logger.debug("Manifest is TAM verified and says TAM is required, updating security database...")
                file = tam_required_file(repository)
                open(file, "w").close()
            if not manifest_required and security_required:
                logger.debug("Manifest is TAM verified and says TAM is *not* required, updating security database...")
                os.unlink(tam_required_file(repository))
        manifest.check_repository_compatibility(operations)
        return manifest

    def check_repository_compatibility(self, operations):
        for operation in operations:
            assert isinstance(operation, self.Operation)
            feature_flags = self.config.get("feature_flags", None)
            if feature_flags is None:
                return
            if operation.value not in feature_flags:
                continue
            requirements = feature_flags[operation.value]
            if "mandatory" in requirements:
                unsupported = set(requirements["mandatory"]) - self.SUPPORTED_REPO_FEATURES
                if unsupported:
                    raise MandatoryFeatureUnsupported(list(unsupported))

    def get_all_mandatory_features(self):
        result = {}
        feature_flags = self.config.get("feature_flags", None)
        if feature_flags is None:
            return result

        for operation, requirements in feature_flags.items():
            if "mandatory" in requirements:
                result[operation] = set(requirements["mandatory"])
        return result

    def write(self):
        from .item import ManifestItem

        if self.key.tam_required:
            self.config["tam_required"] = True
        # self.timestamp needs to be strictly monotonically increasing. Clocks often are not set correctly
        if self.timestamp is None:
            self.timestamp = datetime.now(tz=timezone.utc).isoformat(timespec="microseconds")
        else:
            incremented_ts = self.last_timestamp + timedelta(microseconds=1)
            now_ts = datetime.now(tz=timezone.utc)
            max_ts = max(incremented_ts, now_ts)
            self.timestamp = max_ts.isoformat(timespec="microseconds")
        # include checks for limits as enforced by limited unpacker (used by load())
        assert len(self.archives) <= MAX_ARCHIVES
        assert all(len(name) <= 255 for name in self.archives)
        assert len(self.item_keys) <= 100
        manifest = ManifestItem(
            version=1,
            archives=StableDict(self.archives.get_raw_dict()),
            timestamp=self.timestamp,
            config=StableDict(self.config),
            item_keys=tuple(sorted(self.item_keys)),
        )
        self.tam_verified = True
        data = self.key.pack_and_authenticate_metadata(manifest.as_dict())
        self.id = self.repo_objs.id_hash(data)
        self.repository.put(self.MANIFEST_ID, self.repo_objs.format(self.MANIFEST_ID, {}, data))
